# A model for protein incoherent scattering
# neglecting coherent and maybe some other contributions

import jscatter as js
import numpy as np
import urllib
from jscatter.dynamic import convolve as conv
from jscatter.dynamic import transDiff_w
from jscatter.dynamic import rotDiffusion_w
from jscatter.dynamic import diffusionHarmonicPotential_w

# get pdb structure file with 3rn3 atom coordinates and filter for CA positions in A
# this results in C-alpha representation of protein Ribonuclease A.
url = 'https://files.rcsb.org/download/3RN3.pdb'
pdbtxt = urllib.request.urlopen(url).read().decode('utf8')
CA = [line[31:55].split() for line in pdbtxt.split('\n') if line[13:16].rstrip() == 'CA' and line[:4] == 'ATOM']
# conversion to nm and center of mass
ca = np.array(CA, float) / 10.
cloud = ca - ca.mean(axis=0)
Dtrans = 0.086  # nm**2/ns  for 3rn3 in D2O at 20°C
Drot = 0.0163  # 1/ns
natoms = cloud.shape[0]

# A list of wavevectors and frequencies and a resolution.
# If bgr is zero we have to add it later after the convolution with resolution as constant instrument background.
# As resolution a single q independent (unrealistic) Gaussian is used.
ql = np.r_[0.1, 0.5, 1, 2:15.:2, 20, 30, 40]
w = np.r_[-100:100:0.1]
start = {'s0': 0.5, 'm0': 0, 'a0': 1, 'bgr': 0.00}
resolution = lambda w: js.dynamic.resolution_w(w=w, **start)
res = resolution(w)

# translational diffusion
p = js.grace(3, 0.75)
p.multi(1, 4)
p[0].title(' ' * 50 + 'Inelastic neutron scattering for protein Ribonuclease A', size=2)

iqwRtr = js.dL([conv(transDiff_w(w, q=q, D=Dtrans), res, normB=True) for q in ql])
iqwRt = js.dL([transDiff_w(w, q=q, D=Dtrans) for q in ql])

p[0].plot(iqwRtr, le='$wavevector')
p[0].plot(iqwRt, li=1, sy=0)
p[0].yaxis(min=1e-8, max=1e3, scale='l', ticklabel=['power', 0, 1], label=[r'I(Q,\xw\f{})', 1.5])
p[0].xaxis(min=-149, max=99, label=r'\xw\f{} / 1/ns', charsize=1)
p[0].legend(charsize=1, x=-140, y=1)
p[0].text(r'translational diffusion, one atom', y=80, x=-130, charsize=1.5)
p[0].text(r'Q / nm\S-1', y=2, x=-140, charsize=1.5)

# rotational diffusion
iqwRr = js.dL([rotDiffusion_w(w, q=q, cloud=cloud, Dr=Drot) for q in ql])
iqwRrr = js.dL([conv(iqw, res, normB=True) for iqw in iqwRr])

p[1].plot(iqwRr, li=1, sy=0)
p[1].plot(iqwRrr, le='$wavevector')
p[1].yaxis(min=1e-8, max=1e3, scale='l', ticklabel=['power', 0, 1, 'normal'])
p[1].xaxis(min=-149, max=99, label=r'\xw\f{} / 1/ns', charsize=1)
# p[1].legend()
p[1].text(r'rotational diffusion full cloud', y=50, x=-130, charsize=1.5)

# restricted diffusion in a harmonic local potential
# rmsd defines the size of the harmonic potential
iqwRgr = js.dL([conv(diffusionHarmonicPotential_w(w, q=q, tau=0.15, rmsd=0.3), res, normB=True) for q in ql])
iqwRg = js.dL([diffusionHarmonicPotential_w(w, q=q, tau=0.15, rmsd=0.3) for q in ql])

p[2].plot(iqwRgr, le='$wavevector')
p[2].plot(iqwRg, li=1, sy=0)
p[2].yaxis(min=1e-8, max=1e3, scale='l', ticklabel=['power', 0, 1], label='')
p[2].xaxis(min=-149, max=99, label=r'\xw\f{} / 1/ns', charsize=1)
# p[2].legend()
p[2].text(r'restricted diffusion one atom \n(harmonic)', y=50, x=-130, charsize=1.5)

# amplitudes at w=0 and w=10
p[3].title('amplitudes w=[0, 10]')
p[3].subtitle(r'\xw\f{}>10 restricted diffusion > translational diffusion')
ww = 10
wi = np.abs(w - ww).argmin()
p[3].plot(iqwRtr.wavevector, iqwRtr.Y.array.max(axis=1) * natoms, sy=[1, 0.3, 1], li=[1, 2, 1], le='trans + res')
p[3].plot(iqwRt.wavevector, iqwRt.Y.array.max(axis=1)* natoms, sy=[1, 0.3, 1], li=[2, 2, 1], le='trans ')
p[3].plot(iqwRt.wavevector, iqwRt.Y.array[:, wi]* natoms, sy=[2, 0.3, 1], li=[3, 3, 1], le='trans w=%.2g' % ww)
p[3].plot(iqwRrr.wavevector, iqwRrr.Y.array.max(axis=1), sy=[1, 0.3, 2], li=[1, 2, 2], le='rot + res')
p[3].plot(iqwRr.wavevector, iqwRr.Y.array.max(axis=1), sy=[1, 0.3, 2], li=[2, 2, 2], le='rot')
p[3].plot(iqwRr.wavevector, iqwRr.Y.array[:, wi], sy=[2, 0.3, 2], li=[3, 3, 2], le='rot w=%.2g' % ww)
p[3].plot(iqwRgr.wavevector, iqwRgr.Y.array.max(axis=1)* natoms, sy=[1, 0.3, 3], li=[1, 2, 3], le='restricted + res')
p[3].plot(iqwRg.wavevector, iqwRg.Y.array.max(axis=1)* natoms, sy=[8, 0.3, 3], li=[2, 2, 3], le='restricted')
p[3].plot(iqwRg.wavevector, iqwRg.Y.array[:, wi]* natoms, sy=[3, 0.3, 3], li=[3, 3, 3], le='restricted w=%.2g' % ww)

p[3].yaxis(min=1e-4, max=5e4, scale='l', ticklabel=['power', 0, 1, 'opposite'],
           label=[r'I(Q,\xw\f{}=[0,10])', 1, 'opposite'])
p[3].xaxis(min=0.1, max=50, scale='l', label=r'Q / nm\S-1', charsize=1)
p[3].legend(charsize=0.9, x=2.8, y=0.057)

p[3].text(r'translation', y=8e-1, x=15, rot=331, charsize=1.3, color=1)
p[3].text(r'rotation', y=6e1, x=15, rot=331, charsize=1.3, color=2)
p[3].text(r'harmonic', y=1e-1, x=15, rot=331, charsize=1.3, color=3)
p[3].text(r'\xw\f{}=10', y=1e-2, x=0.2, rot=30, charsize=1.5, color=1)
p[3].line(0.17, 5e-2, 0.17, 8e-3, 5, arrow=2, color=1)

p[3].text(r'resolution', y=2212, x=0.5, rot=0, charsize=1.5, color=4)

p[3].line(0.5, 100, 0.5, 25, 5, arrow=2, color=4)
p[3].line(0.5, 1500, 0.5, 300, 5, arrow=2, color=4)

p.save('inelasticNeutronScattering.png', size=(6, 1.5), dpi=400)

# all together in a combined model  ----------------------------
start = {'s0': 0.5, 'm0': 0, 'a0': 1, 'bgr': 0.00}
resolution = lambda w: js.dynamic.resolution_w(w=w, **start)


def transrotsurfModel(w, q, Dt, Dr, exR, tau, rmsd):
    """
    A model for trans/rot diffusion with a partial local restricted diffusion at the protein surface.

    See Fast internal dynamics in alcohol dehydrogenase The Journal of Chemical Physics 143, 075101 (2015);
    https://doi.org/10.1063/1.4928512

    Parameters
    ----------
    w   frequencies
    q   wavevector
    Dt  translational diffusion
    Dr  rotational diffusion
    exR outside this radius additional restricted diffusion with t0 u0
    tau  correlation time
    rmsd  Root mean square displacement Ds=u0**2/t0

    Returns
    -------

    """
    natoms = cloud.shape[0]
    trans = transDiff_w(w, q, Dt)
    trans.Y = trans.Y * natoms  # natoms contribute
    rot = rotDiffusion_w(w, q, cloud, Dr)  # already includes natoms
    fsurf = ((cloud ** 2).sum(axis=1) ** 0.5 > exR).sum()  # fraction of natoms close to surface
    loc = diffusionHarmonicPotential_w(w, q, tau, rmsd)
    # only fsurf atoms at surface contributes to local restricted diffusion, others elastic
    loc.Y = js.dynamic.elastic_w(w).Y * (natoms - fsurf) + fsurf * loc.Y

    final = conv(trans, rot)
    final = conv(final, loc)
    final.setattr(rot, 'rot_')
    final.setattr(loc, 'loc_')

    res = resolution(w)
    finalres = conv(final, res, normB=True)

    # finalres.Y+=0.0073  # background ?
    finalres.q = q
    finalres.fsurf = fsurf
    return finalres


ql = np.r_[0.1, 0.5, 1, 2, 4, 6, 10, 20]
p = js.grace(1, 1)
p.title('Protein incoherent scattering')
p.subtitle('Ribonuclease A')
iqwR = js.dL([transrotsurfModel(w, q=q, Dt=Dtrans, Dr=Drot, exR=0, tau=0.15, rmsd=0.3) for q in ql])
p.plot(iqwR[0], sy=0, li=[1, 2, 1], le='exR=0')
p.plot(iqwR[1:], sy=0, li=[1, 2, 1])
iqwR = js.dL([transrotsurfModel(w, q=q, Dt=Dtrans, Dr=Drot, exR=1, tau=0.15, rmsd=0.3) for q in ql])
p.plot(iqwR[0], sy=0, li=[1, 2, 2], le='exR=1')
p.plot(iqwR[1:], sy=0, li=[1, 2, 2])
iqwR = js.dL([transrotsurfModel(w, q=q, Dt=Dtrans, Dr=Drot, exR=1.4, tau=0.15, rmsd=0.3) for q in ql])
p.plot(iqwR[0], sy=0, li=[3, 3, 3], le='exR=1.4')
p.plot(iqwR[1:], sy=0, li=[3, 3, 3])
p.yaxis(min=1e0, max=2e5, label=r'I(q,\xw\f{})', scale='l', ticklabel=['power', 0, 1])
p.xaxis(min=0.1, max=100, label=r'\xw\f{} / ns\S-1', scale='l')
p.legend(x=0.2, y=1e1)

p[0].line(1, 5e2, 1, 5e1, 2, arrow=1)
p.text('resolution', x=0.9, y=7e1, rot=90)
p.text(r'q=2 nm\S-1', x=0.2, y=1e5, rot=0)
p.text(r'q=20 nm\S-1', x=0.2, y=8e2, rot=0)
p.text(r'q=3 nm\S-1', x=2, y=5e4, rot=0)
p.text(r'q=0.1 nm\S-1', x=2.2, y=1.5e2, rot=0)

p[0].line(0.17, 9e4, 0.17, 1e3, 2, arrow=2)
p[0].line(2, 1.5e2, 2, 4e4, 2, arrow=2)
p.text(r'Variation \nfixed/harmonic protons', x=1.2, y=4.44, rot=0)
p[0].line(7, 5.3, 12, 55, arrow=2)

p.save('Ribonuclease_inelasticNeutronScattering.png', size=(4, 4), dpi=150)



import jscatter as js

# Comparison of different structure factors

R = 3
phi = 0.1
length = 4
depth = 1
q = js.loglist(0.01, 5, 200)

PY = js.sf.PercusYevick(q, R, eta=phi)
shsp = js.sf.stickyHardSphere(q, R, length, depth, phi=phi)
ahs = js.sf.adhesiveHardSphere(q, R, length, depth, eta=phi)
rmsa = js.sf.RMSA(q, R, scl=length, gamma=depth, eta=phi)
p = js.grace()
p.plot(PY, symbol=0, line=[1, 2, 1], legend='PercusYevick')
p.plot(shsp, symbol=0, line=[1, 2, 2], legend='sticky hard sphere')
p.plot(ahs, symbol=0, line=[1, 2, 4], legend='adhesive hard sphere')
p.plot(rmsa, symbol=0, line=[1, 2, 3], legend='RMSA')
p.yaxis(min=0.0, max=1.5, label='S(Q)', charsize=1.5)
p.legend(x=3, y=0.8)
p.xaxis(min=0, max=5, label=r'Q / nm\S-1')
p.title('Comparison of different structure factors')
p.save('Comparisonofdifferentstructurefactors.png')



# -*- coding: utf-8 -*-
# written by Ralf Biehl at the Forschungszentrum Jülich ,
# Jülich Center for Neutron Science 1 and Institute of Complex Systems 1
#    Jscatter is a program to read, analyse and plot data
#    Copyright (C) 2015-2024  Ralf Biehl
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

import os
import numbers

import numpy as np
import scipy.special as special
from scipy.special import roots_legendre

from ..dataarray import dataArray as dA

_path_ = os.path.realpath(os.path.dirname(__file__))


def _cached_roots_legendre(n):
    """
    Cache roots_legendre results to speed up calls of the fixed_quad function.
    """
    # scipy.integrate.quadrature
    if n in _cached_roots_legendre.cache:
        return _cached_roots_legendre.cache[n]

    _cached_roots_legendre.cache[n] = roots_legendre(n)
    return _cached_roots_legendre.cache[n]


_cached_roots_legendre.cache = dict()


def fa_sphere(qr):
    """
    scattering amplitude sphere with catching the zero
    qr is array dim 1

    """
    fa=np.ones(qr.shape)
    qr0 = (qr!=0)
    fa[qr0] = 3 / qr[qr0] ** 3 * (np.sin(qr[qr0]) - qr[qr0] * np.cos(qr[qr0]))

    return fa


def fa_coil(qrg):
    """
    qrg is array dim 1

    We use the root of the Debye function here

    The fa(x)=(1-exp(-x))/x given in some papers result in the wrong
    high Q limit of q**-4 as it is only valid for small qrg<<1

    """
    fa = np.ones(qrg.shape)
    x = qrg**2
    # root of Debye function, which is a always positive function without extreme values
    fa[x != 0] =(2*(np.exp(-x)-1+x))**0.5 / x

    return fa


def fa_disc(q, R, D, angle):
    """
    Multishell disc form factor amplitude, save for q=0 and q<0 result is zero

    q : wavevectors array [q,None]
    D : thickness of discs along axis, array  [None,D]
    R : Radii of discs, array [None,R]
    angle : angle between axis and scattering vector q in rad array [None,None,angle]

    q<0 result is zero needed in ellipsoidFilledCylinder

    Return
    ------
    array with fa(q,angle)

    """
    result = np.zeros((q.shape[0], D.shape[1], angle.shape[2]))

    sina = np.sin(angle)
    cosa = np.cos(angle)
    # set sin(a) to 1 for angle = 0
    sina[angle == 0] = 1

    if D[0, 0, 0] > 0 and R[0, 0, 0] > 0:
        fq0 = 2. * np.pi * R ** 2 * D

        def fqq(q):
            return fq0 * special.j1(q * R * sina) / (q * R * sina) * np.sinc(q * D / 2. * cosa / np.pi)

    elif R[0, 0, 0] > 0:
        fq0 = 2. * np.pi * R ** 2 * 1

        def fqq(q):
            return fq0 * special.j1(q * R * sina) / (q * R * sina)

    elif D[0, 0, 0] > 0:
        fq0 = 2. * D

        def fqq(q):
            return fq0 * np.sinc(q * D / 2. * cosa / np.pi)

    result[np.where(q > 0)[0]] = fqq(q[np.where(q > 0)[0]])
    result[np.where(q == 0)[0]] = fq0 * 0.5
    return result


def fq_disc(QQ, R, D, angle, dSLDs):
    """
    Multishell disc form factor, save for q=0 and q<0 result is zero

    q : array
        wavevectors
    D : array
        thickness of discs along axis, array
    R : array
        Radii of discs, array
    angle : array
        angles between axis and scattering vector q in rad

    q<0 result is zero needed in ellipsoidFilledCylinder

    Return
    ------
    array with f(q)={ fa(q,angle)**2 as 2d array [q,angles]

    """
    # distribute q,R,D,angle and sld  along arrays dimensions
    Pc = dSLDs[None, :, None] * fa_disc(QQ[:, None, None], R[None, :, None], D[None, :, None], angle[None, None, :])
    if len(R) > 1:  # subtract lower integration boundary
        #  innermost R not included
        Pc[:, 1:] = (Pc[:, 1:] - dSLDs[None, 1:, None] *
                     fa_disc(QQ[:, None, None], R[None, :-1, None], D[None, :-1, None], angle[None, None, :]))
    # square it for fq
    Pc2 = Pc.sum(axis=1) ** 2

    return Pc2


def _thindisc(q, ri, ro, z, angle):
    # z element of a thin discs for integration along z,
    # equ.  11 (sin(theta) should be outside root)
    # in Kaya J. Appl. Cryst. (2004). 37, 508-509 DOI: 10.1107/S0021889804005709
    #
    # q scattering vector; q = QQ[:, None, None],
    # angle : angles theta between Z axis and q vector;   angle = angle[None, None, :]
    # ri, ro inner and outer radius of the disc along the Z axis;  ri = rimradii[None, :, None],
    # z height along Z axis for integration along Z for pssitive z but negative included using symmetry

    qcos = (q * np.cos(angle))[None, :]
    sina = np.sin(angle)
    sina[angle == 0] = 1
    qsin = (q * sina)[None, :]

    outside = ro * special.j1(qsin * ro)
    inside = ri * special.j1(qsin * ri)
    return 4 * np.pi / q[None, :] / sina[None, :] * np.cos(qcos * z) * (outside - inside)


def fa_rim(q, rr, R, D, angle, p=1, n=21):
    # fa of the curved rim around a multi shell disc with disc axis along z axis and radius R
    # q : scattering vectors
    # R : (float) outside radius of the disc = inside radius of rim
    # D :Shell thickness (is full thickness)
    # rr : rim radii at equator (xy plane) also rim radii in z direction at disc outer radius,
    #      each rim border is not necessarily continuous to disc shell
    #      for equal rim thickness of consecutive layers the corners remain, only if *all* inner rims =0 this vanishes
    # angle : angles between disc axis and q
    p2 = p*2  # exponent that describes curved rim like superball

    # integration knot points and weights for Gaussian integration
    x, w = _cached_roots_legendre(n)
    x = np.real(x)

    # integration along z over rim as discs to get curved rim
    # rout describes outer shape of the rim , inside is flat with radius R
    uplimit = D.max()/2  # outer thickness = max thickness in z direction
    lowlimit = 0  # use symmetry along z=0 in integration
    # supporting points for integration variable dz with weights w
    z = ((uplimit - lowlimit) * (x[:, None, None, None] + 1) / 2.0 + lowlimit)  # first axis for z
    # first index is z integrations with w, second last for subtracting shells dependent on rr, last for angular average
    result = np.zeros((z.shape[0], q.shape[0], rr.shape[1], angle.shape[2]))

    # outer radii on rim dependent on z and p
    rp = rr[None, :]**p2 * (1 - (z/(D[None, :] * 0.5))**p2)
    rp[rp < 0] = 0  # some are negative (z>rr), catch them
    rout = R[None, :] + rp**(1/p2)
    # inner radii all equal
    rin = R[None, :] * np.ones_like(rp)
    # rim areas along z
    rim0 = 4*np.pi * (rout**2 - R[None, :]**2)

    # calc values at knots for z integration
    result[:, np.where(q > 0)[0]] = (uplimit - lowlimit) / 2.0 * _thindisc(q[np.where(q > 0)[0]], rin, rout, z, angle)
    result[:, np.where(q == 0)[0]] = (uplimit - lowlimit) / 2.0 * rim0 * 0.5

    # multiply by weight and sum over weights to integrate along z
    return (result * w[:, None, None, None]).sum(axis=0)


def fq_rimdisc(QQ, D, R, angle, rimradii, dSLDs, drSLDs, p, nrim):
    # formfactor of a disc with outer half spherical rim for p=1, rectangular rim for p>10
    # q : array, scattering vectors
    # R : float,  radius of the disc
    # D : array, thickness of the discs
    # rimdradii : array, rim radii, not necessarily the same as D
    # p : float, curvature of rim similar to superball
    # angle : array, angles between Z axis and disc axis to average
    # dSLDs : array, contrast for shells,
    # drSLDs : array, contrast for rim shells,
    # nrim:  integration steps for rim
    # return:   arrays for all Q (axis 0) and all shells (axis 1) and all angles (axis 2)

    R_ = np.ones_like(D)[None, :, None] * np.atleast_1d(R)[None, None, :]  # for discs we need a list of equal R

    # calc outer discs,
    # second dimension is for consecutive discs, last dim for integration over R and angular average from calling fkt
    Pc = dSLDs[None, :, None] * fa_disc(QQ[:, None, None], R_, D[None, :, None], angle[None, None, :])

    if rimradii is not None and np.any(rimradii > 0):
        # we have a rim
        # calc cap contribution
        Pcap = drSLDs[None, :, None] * fa_rim(q=QQ[:, None, None],
                                              rr=rimradii[None, :, None],
                                              R=R_,
                                              D=D[None, :, None],
                                              angle=angle[None, None, :], p=p, n=nrim)

    if len(D) > 1:
        # subtract inner cylinders that shell remains
        #  innermost is not subtracted
        Pc[:, 1:] = (Pc[:, 1:] - dSLDs[None, 1:, None] *
                     fa_disc(QQ[:, None, None], R_[:, :-1, :], D[None, :-1, None], angle[None, None, :]))
        if rimradii is not None and np.any(rimradii > 0):
            # subtract inside caps
            Pcap[:, 1:] = Pcap[:, 1:] - drSLDs[None, 1:, None] * fa_rim(q=QQ[:, None, None],
                                                                        rr=rimradii[None, :-1, None],
                                                                        R=R_[:, :-1, :],
                                                                        D=D[None, :-1, None],
                                                                        angle=angle[None, None, :], p=p, n=nrim)

    # sum up all cylinder shells with axis=1
    if rimradii is not None and np.any(rimradii > 0):
        # this avoids the infinite thin disc to be added
        if np.all(R > 0):
            Pcs = (Pc + Pcap).sum(axis=1)
        else:
            Pcs = Pcap.sum(axis=1)
    else:
        # cylinder without cap
        Pcs = Pc.sum(axis=1)

    # return scattering amplitude**2
    return Pcs**2


def fa_cylinder(q, r, L, angle):
    """
    cylinder form factor amplitude, save for q=0 and q<0 result is zero

    q : wavevectors
    r : shell thickness , a list or array !!
    L : length of cylinder, L=0 is infinitely long cylinder
    angle : angle between axis and scattering vector q in rad

    q<0 result is zero needed in ellipsoidFilledCylinder

    """
    # deal with possible zero in q
    if isinstance(q, numbers.Number):
        q = np.r_[q]
    result = np.zeros((len(q), len(r)))
    if angle != 0:
        sina = np.sin(angle)
        cosa = np.cos(angle)
    else:
        sina = 1
        cosa = 1
    if L > 0 and r[0] > 0:
        fq0 = 2. * np.pi * r ** 2 * L
        fqq = lambda qq: fq0 * special.j1(qq[:, None] * r * sina) / (qq[:, None] * r * sina) * \
                         np.sinc(qq[:, None] * L / 2. * cosa / np.pi)
    elif r[0] > 0:
        fq0 = 2. * np.pi * r ** 2 * 1
        fqq = lambda qq: fq0 * special.j1(qq[:, None] * r * sina) / (qq[:, None] * r * sina)
    elif L > 0:
        fq0 = 2. * L
        fqq = lambda qq: fq0 * np.sinc(qq[:, None] * L / 2. * cosa / np.pi)
    result[np.where(q > 0)[0], :] = fqq(q[np.where(q > 0)])
    result[np.where(q == 0)[0], :] = fq0 * 0.5

    return result


def fa_cylindercap(q, r, L, angle, h, n=21):
    # fa of a cylinder cap as spherical end of a cylinder with height h
    # Equ 1 in Kaya & Souza  J. Appl. Cryst. (2004). 37, 508±509  DOI: 10.1107/S0021889804005709
    # integrate by fixed Gaussian at positions t and weights w
    j1 = special.j1

    # integration knot points and weights for Gaussian integration
    x, w = _cached_roots_legendre(n)
    x = np.real(x)

    if isinstance(q, numbers.Number):
        q = np.r_[q]
    if angle != 0:
        sina = np.sin(angle)
        cosa = np.cos(angle)
    else:
        sina = 1
        cosa = 1

    R = (h ** 2 + r ** 2) ** 0.5

    # integration limits
    lowlimit = -h / R
    uplimit = 1
    t = ((uplimit - lowlimit) * (x[:, None, None] + 1) / 2.0 + lowlimit)  # first axis for x
    result = np.zeros((len(t), len(q), len(r)))

    def cap(q):
        return 4 * np.pi * r ** 3 * np.cos(q[:, None] * cosa * (r * t + h + L / 2)) * \
                    (1 - t ** 2) * (j1(q[:, None] * r * sina * (1 - t ** 2) ** 0.5)) / \
                    (q[:, None] * r * sina * (1 - t ** 2) ** 0.5)
    cap0 = 4 * np.pi * r ** 3 * (1 - t ** 2)

    # calc values at knots
    result[:, np.where(q > 0)[0], :] = (uplimit - lowlimit) / 2.0 * cap(q[np.where(q > 0)])
    result[:, np.where(q == 0)[0], :] = (uplimit - lowlimit) / 2.0 * cap0 * 0.5

    # multiply by weight and sum over weights to integrate
    return (result * w[:, None, None]).sum(axis=0)


def fa_capedcylinder(QQ0, r, L, angle, h, dSLDs, ncap):
    # formfactor amplitude of a cylinder with orientation alpha and cap
    # outer integration boundary r
    # L cylinder length, angle orientation
    # h center of spherical cap relative to cylinder end.
    #   See picture in formfactor.composed.multiShellCylinder
    # dSLDs contrast for multi shells,
    # ncap integration steps for cap
    # the functions _fa_ return arrays for all Q (axis 0) and all shells (axis 1)

    # calc outer cylinders
    Pc = dSLDs * fa_cylinder(QQ0, r, L, angle)
    if h is not None and np.all(r > 0):
        # calc cap contribution
        Pcap = dSLDs * fa_cylindercap(QQ0, r, L, angle, h, ncap)

    if len(r) > 1:
        # subtract inner cylinders that shell remains
        #  innermost with r==0 is not subtracted
        Pc[:, 1:] = Pc[:, 1:] - dSLDs[1:] * fa_cylinder(QQ0, r[:-1], L, angle)
        if h is not None and np.all(r > 0):
            # calc cap contribution
            Pcap[:, 1:] = Pcap[:, 1:] - dSLDs[1:] * fa_cylindercap(QQ0, r[:-1], L, angle, h, ncap)

    # sum up all cylinder shells with axis=1
    if h is not None and np.all(r > 0):
        # this avoids the infinite thin disc to be added
        if L > 0:
            Pcs = (Pc + Pcap).sum(axis=1)
        else:
            Pcs = Pcap.sum(axis=1)
    else:
        # cylinder without cap
        Pcs = Pc.sum(axis=1)

    # return scattering amplitude
    return Pcs


def fq_capedcylinder(QQ, r, L, angle, h, dSLDs, ncap):
    # calc scattering amplitude and square it for formfactor
    # include zero for forward scattering
    fa = fa_capedcylinder(np.r_[0, QQ], r, L, angle, h, dSLDs, ncap)
    result = dA(np.c_[QQ, fa[1:]**2].T)
    # store the forward scattering
    result.I0 = fa[0]**2
    return result


def fq_cuboid(q, p, t, a, b, c):
    """
    Scattering of cuboid with orientation pt

    Parameters
    ----------
    q : array wavevector
    p : array angle phi
    t: array angle theta
    a,b,c : float edge length of cuboid

    """
    pi2 = np.pi * 2
    fa = (np.sinc(q * a * np.sin(t[:, None]) * np.cos(p[:, None]) / pi2) *
          np.sinc(q * b * np.sin(t[:, None]) * np.sin(p[:, None]) / pi2) *
          np.sinc(q * c * np.cos(t[:, None]) / pi2)) ** 2 * np.sin(t[:, None])
    return fa


def fq_prism(points, Q, R, H):
    """
    Equal sided prism width edge length R of height H

    The height is along Z-axis. The prism rectangular basis is parallel to XZ-plane,
    the triangular plane is parallel to XY-plane. See [1]_ SI *The form factor of a prism*.

    Parameters
    ----------
    points : 3xN array
        q direction on unit sphere
    Q 1xM array
        Q values
    R : float
        2R is edge length
    H : float
        Prism height in Z direction

    Returns
    -------
        array

    References
    ----------
    .. [1] DNA-Nanoparticle Superlattices Formed From Anisotropic Building Blocks
          Jones et al.
          Nature Materials 9, 913–917 (2010), doi: 10.1038/nmat2870

    """
    qx, qy, qz = points.T[:, :, None] * Q[None, None, :]
    sq3 = np.sqrt(3)
    fa_prism = 2*sq3*np.exp(-1j*qy*R/sq3)*H / (qx*(qx**2-3*qy**2)) * \
               (qx*np.exp(1j*qy*R*sq3) - qx*np.cos(qx*R) - 1j*sq3*qy*np.sin(qx*R)) * \
               np.sinc(qz*H/2)

    return np.real(fa_prism * np.conj(fa_prism))


def gauss(x, a, s):
    # Gaussian normalized to have Integral s
    return np.exp(-0.5 * (x - a) ** 2 / s ** 2) / np.sqrt(2 * np.pi)


def PHI(u):
    return 3*(np.sin(u) - u*np.cos(u)) / u**3


def fq_triellipsoid(q, p, t, Ra, Rb, Rc):
    qx = q * np.sin(t[:, None]) * np.cos(p[:, None])
    qy = q * np.sin(t[:, None]) * np.sin(p[:, None])
    qz = q * np.cos(t[:, None])
    # J. Appl. Cryst. (2020). 53, 1387-1391  https://doi.org/10.1107/S1600576720010389
    qr = (qx**2 + (Rb/Ra*qy)**2 + (Rc/Ra*qz)**2)**0.5
    return PHI(qr*Ra)**2 * np.sin(t[:, None])




import jscatter as js

# first get and create the biological unit ('.pdb1') of alcohol dehydrogenase (tetramer, 144 kDa)
adh = js.bio.fetch_pdb('4w6z.pdb1')
# the 2 dimers in are in model 1 and 2 and need to be merged into one.
adhmerged = js.bio.mergePDBModel(adh)

# Ribonuclease A and Phosphoglycerate kinase are monomers and can be used without modifications.
# 3pgk has a Mg atom that is misinterpreted (as M), to use it we add this
vdwradii = {'M': js.data.vdwradii['Mg'] * 10}  # in A

p = js.grace(1, 1.4)
p.multi(3, 2, vgap=0, hgap=0)
for c, pdbid in enumerate(['3rn3', '3pgk', adhmerged]):
    # load from pdb id, scatteringUniverse adds hydrogens automatically
    uni = js.bio.scatteringUniverse(pdbid, vdwradii=vdwradii)
    uni.setSolvent('1d2o1')
    uni.qlist = js.loglist(0.1, 9.9, 200)
    u = uni.select_atoms("protein")
    ur = u.residues

    S = js.bio.nscatIntUniv(u)
    Sx = js.bio.xscatIntUniv(u)

    # use an averaging in cubes filled with the atoms, cubesize approximates residue size
    Scu = js.bio.nscatIntUniv(u, cubesize=0.6)
    Sxcu = js.bio.xscatIntUniv(u, cubesize=0.6)

    # use now the precalculated formfactors on residue level coarse graining
    uni.explicitResidueFormFactorAmpl = False  # default
    Sr = js.bio.nscatIntUniv(ur)
    Sxr = js.bio.xscatIntUniv(ur)

    # calc residue formfactors explicit (not precalculated)
    # useful for changes of residue deuteration or backbone N-H exchange of IDP
    uni.explicitResidueFormFactorAmpl = True
    Ser = js.bio.nscatIntUniv(ur)
    Sxer = js.bio.xscatIntUniv(ur)

    # create a C-alpha pdb file and then the Ca-only universe for calculation
    ca = uni.select_atoms('name CA')
    ca.write('pdb_ca.pdb')
    # addHydrogen=False prevents addition of 4H per C atom
    unica = js.bio.scatteringUniverse('pdb_ca.pdb', addHydrogen=False)
    # To use precalculated residue formfactors explicit... should be False
    unica.explicitResidueFormFactorAmpl = False
    unica.setSolvent('1d2o1')
    unica.qlist = js.loglist(0.1, 10, 200)
    uca = unica.residues
    Sca = js.bio.nscatIntUniv(uca, getVolume='now')
    Sxca = js.bio.xscatIntUniv(uca)

    p[2 * c].plot(S, li=[1, 2, 1], sy=0, le='atomic')
    p[2 * c].plot(Scu, li=[1, 2, 5], sy=0, le='atomic in cubes')
    p[2 * c].plot(Sr, li=[1, 2, 2], sy=0, le='res pre')
    p[2 * c].plot(Ser, li=[3, 2, 3], sy=0, le='res ex')
    p[2 * c].plot(Sca, li=[1, 2, 4], sy=0, le='Ca model')
    p[2 * c].legend(x=1, y=8e-3, charsize=0.8)
    p[2 * c].text(x=0.15, y=1e-7, charsize=1, string=pdbid)

    p[2 * c + 1].plot(Sx, li=[1, 2, 1], sy=0, le='atomic')
    p[2 * c + 1].plot(Sxcu, li=[1, 2, 5], sy=0, le='atomic in cubes')
    p[2 * c + 1].plot(Sxr, li=[1, 2, 2], sy=0, le='res pre')
    p[2 * c + 1].plot(Sxer, li=[3, 2, 3], sy=0, le='res ex')
    p[2 * c + 1].plot(Sxca, li=[1, 2, 4], sy=0, le='Ca model')
    p[2 * c + 1].legend(x=1, y=8e-3, charsize=0.8)
    p[2 * c + 1].text(x=0.15, y=1e-7, charsize=1, string=pdbid)

    p[2 * c].xaxis(label='', ticklabel=False, scale='log', min=1e-1, max=9.9)
    p[2 * c + 1].xaxis(label='', ticklabel=False, scale='log', min=1e-1, max=9.9)
    p[2 * c].yaxis(label='F(Q)', ticklabel='power', scale='log', min=3e-8, max=8e-3)
    p[2 * c + 1].yaxis(ticklabel=False, scale='log', min=3e-8, max=8e-3)

p[2 * c].xaxis(label=r'Q / nm\S-1', ticklabel=True, scale='log', min=1e-1, max=9.9)
p[2 * c + 1].xaxis(label=r'Q / nm\S-1', ticklabel=True, scale='log', min=1e-1, max=9.9)
p[0].subtitle('neutron scattering')
p[1].subtitle('Xray scattering')
p[0].title(' ' * 30 + 'Comparison of formfactors with different resolution')
# p.save(js.examples.imagepath+r'/uniformfactors.jpg', size=(700/300, 1000/300))




# -*- coding: utf-8 -*-
# written by Ralf Biehl at the Forschungszentrum Jülich ,
# Jülich Center for Neutron Science 1 and Institute of Complex Systems 1
#    Jscatter is a program to read, analyse and plot data
#    Copyright (C) 2015-2024  Ralf Biehl
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

"""
Some examples to show how to use Jscatter.

Functions show the code or run the examples in example directory.



"""

import os
import glob
import re
import platform
import subprocess
import tempfile
import numbers

from .. import graceplot
from .particlesWithInteraction import particlesWithInteraction

platformname = platform.uname()[0]

_path_ = os.path.realpath(os.path.dirname(__file__))
_expath = _path_
datapath = os.path.join(_expath, 'exampleData')
imagepath = os.path.join(_expath, 'images')


def _runsInNotebook():
    """
    Test if working in terminal or notebook

    Returns
    -------
        bool

    """
    try:
        ipy_str = str(type(get_ipython()))
        if 'zmqshell' in ipy_str:
            return 'jupyter'
        if 'terminal' in ipy_str:
            return False  # 'ipython'
    except:
        return False  # 'terminal'


isNotebook = _runsInNotebook()


def _multiple_replace(string, rep_dict):
    # from https://stackoverflow.com/questions/6116978/how-to-replace-multiple-substrings-of-a-string
    pattern = re.compile("|".join([re.escape(k) for k in sorted(rep_dict, key=len, reverse=True)]), flags=re.DOTALL)
    return pattern.sub(lambda x: rep_dict[x.group(0)], string)


def _multiple_replace2(string, rep_dict):
    # from https://stackoverflow.com/questions/6116978/how-to-replace-multiple-substrings-of-a-string
    for k, v in rep_dict.items():
        pattern = re.compile(k, flags=re.DOTALL)
        string = pattern.sub(v, string)
    return string


mplotreplace = {'.grace(': '.mplot(',
                '.plot(': '.Plot(',
                '.save(': '.Save(',
                '.xaxis(': '.Xaxis(',
                '.yaxis(': '.Yaxis(',
                '.legend(': '.Legend(',
                '.clear(': '.Clear(',
                '.title(': '.Title(',
                '.subtitle(': '.Subtitle(',
                '.multi(': '.Multi(',
                '.agr': '.png',
                '.text': '.Text',
                "(r'": "('",
                '.line': '.Arrow',
                r'\xp': 'pi', r'\N': '',
                'new_graph': 'Subplot',
                }
headlessreplace = {'fig.show()': 'fig.savefig("lastopenedplots_0.png")',
                   'fig0.show()': 'fig.savefig("lastopenedplots_0.png")',
                   'fig1.show()': 'fig.savefig("lastopenedplots_1.png")',
                   'fig2.show()': 'fig.savefig("lastopenedplots_2.png")',
                   'fig3.show()': 'fig.savefig("lastopenedplots_3.png")',
                   'fig4.show()': 'fig.savefig("lastopenedplots_4.png")',
                   'fig5.show()': 'fig.savefig("lastopenedplots_5.png")',
                   'pyplot.show': 'js.mpl.show',
                   }


def _getexamples():
    exfiles = glob.glob(_expath + '/example_*.py')
    return sorted(exfiles)


def showExampleList():
    """
    Show an updated list of all examples.

    Currently, availible examples ::

    """
    print('Example path : ', _expath)
    exfiles = _getexamples()
    for i, ff in enumerate(exfiles):
        print(i, '  ', os.path.basename(ff))


# string with example filenames for documentation appended to __doc__
_examplelist=''.join(['      {0:}  {1:} \n'.format(i, os.path.basename(ff)) for i, ff in enumerate(_getexamples())])
showExampleList.__doc__ += _examplelist + '\n'


def runExample(example, usempl=False):
    """
    Run example by number or name.

    Example and saved files go to JscatterExamples/example_name in the current working directory.

    Parameters
    ----------
    example: string,int
        Filename or number of the example to run
    usempl : bool, default False
        For using mpl set to True

    Examples
    -------- ::

     import jscatter as js
     js.examples.runExample('daily_use')

    """
    if not graceplot.GraceIsInstalled or isNotebook:
        usempl = True

    exfiles = _getexamples()
    if isinstance(example, numbers.Integral):
        example = exfiles[example]
        print('-----------------------------------')
        print(example)
    else:
        for ef in exfiles:
            if example in ef:
                example = ef
                break

    cwd = os.getcwd()
    # try and in anny case go back to cwd
    try:
        # change path to write into subfolder
        head, exname = os.path.split(os.path.splitext(example)[0])
        temppath = os.path.join(cwd, 'jscatterExamples', exname)
        if not os.path.exists(temppath):
            os.makedirs(temppath)
        os.chdir(temppath)
        with open(example) as f:
            if usempl:
                if 'grace' in example or 'makeJ' in example:
                    print('This example is not fo r mpl.')
                    return
                file = f.read()
                file = _multiple_replace2(file, {r'\.grace\(\d\.?\d?,\s?\d\.?\d?\)': ".mplot()"})
                file = _multiple_replace(file, mplotreplace)
            else:
                file=f.read()
            if graceplot._headless:
                print('headless')
                file = _multiple_replace(file, headlessreplace)
                # append this to close all opened mpl figures
                file+='\n\njs.mpl.close("all")\n\n'

        # now do it
        exec(file, {})
    finally:
        os.chdir(cwd)

    # write the executed file to the subfolder
    with open(os.path.join(temppath, exname+'.py'), 'w') as out_file:
        out_file.writelines(file)

    return


def showExample(example='.', usempl=False):
    """
    Opens example in default editor.

    Parameters
    ----------
    example : string, int
        Filename or number.
        If '.' the folder with the examples is opened.
    usempl : bool, default False
        For using mpl set to True.
        Then mpl examples are shown.

    """
    if not graceplot.GraceIsInstalled:
        usempl = True

    exfiles = _getexamples()
    if isinstance(example, numbers.Integral):
        example = exfiles[example]
        print('-----------------------------------')
        print(example)
    else:
        for ef in exfiles:
            if example in ef:
                example = ef
                break
    examplepath = example

    if usempl:
        if 'grace' in example:
            print('This example is not for mpl.', example)
            return
        with tempfile.NamedTemporaryFile(suffix=".txt", prefix=example, delete=False) as tf:
            with open(examplepath) as f:
                file = f.read()
            newfile = _multiple_replace(file, mplotreplace)
            if platformname == 'Darwin':
                tf.write(newfile.encode())
                tf.flush()
                examplepath = tf.name
                # open is broken but seem to be fixed in newest MACOS versions
                subprocess.call(('open',))
            elif platformname == 'Windows':
                tf.write(newfile.replace('\n', '\r\n').encode())
                tf.flush()
                examplepath = tf.name
                # noinspection PyUnresolvedReferences
                os.startfile(examplepath)
            else:
                tf.write(newfile.encode())
                tf.flush()
                examplepath = tf.name
                subprocess.call(('xdg-open', examplepath))
            tf.close()
            print(examplepath)
    else:
        print(examplepath)
        if platformname == 'Darwin':
            # open is broken but seem to be fixed in newest MACOS versions
            subprocess.call(('open',))
        elif platformname == 'Windows':
            # This should not happen
            pass
        else:
            subprocess.call(('xdg-open', examplepath))


def runAll(start=None, end=None):
    """
    Run all examples ( Maybe needs a bit of time ) .

    Example and saved files go to JscatterExamples/example_name in the current working directory.

    Parameters
    ----------
    start,end : int
        First and last example to run

    """
    exfiles = _getexamples()
    for ff in exfiles[start:end]:
        print('----------------------------------')
        print(os.path.basename(ff))
        runExample(ff)



# -*- coding: utf-8 -*-
# written by Ralf Biehl at the Forschungszentrum Jülich ,
# Jülich Center for Neutron Science 1 and Institute of Complex Systems 1
#    Jscatter is a program to read, analyse and plot data
#    Copyright (C) 2015-2024  Ralf Biehl
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

r"""
**Particle solution**

The scattering intensity of isotropic particles in solution with particle concentration :math:`c_p`
and structure factor :math:`S(q)` (:math:`S(q)=1` for non-interacting particles) is

.. math:: I(q)= c_p I_p(q) S(s) = c_p I_0 F(q) S(q)

In this module the scattering intensity :math:`I_p(q)` of a single particle with
real scattering length densities is calculated in units :math:`nm^2=10^{-14} cm^2`.
For the structure factor :math:`S(q)` see :ref:`structurefactor (sf)`.

If the scattering length density is not defined as e.g. for beaucage model
the normalized particle form factor :math:`F(q)` with :math:`F(q=0)=1` is calculated.

Conversion of single particle scattering :math:`I_p(q)` to particle in solution
(units :math:`\frac{1}{cm}` with :math:`c_p` in mol/liter) is
:math:`I_{[1/cm]}(q)=N_A \frac{c_p}{1000} 10^{-14} I_{p,[nm^2]}(q)`.

.. collapse:: Definition of Particle Formfactors

    The particle formfactor is  (:math:`\hat{F} ; normalized`)

    .. math:: F(q) &= F_a(q)F^*_a(q)=|F_a(q)|^2 \\
              \hat{F}(q) &= \hat{F_a}(q)\hat{F^*_a}(q)=|\hat{F_a}(q)|^2

    and particle scattering amplitude

    .. math:: F_a(q) &= \int_V b(r) e^{iqr} \mathrm{d}r  = \sum_N b_i e^{iqr} \\
              \hat{F_a}(q) &= \int_V b(r) e^{iqr} \mathrm{d}r  / \int_V b(r) \mathrm{d}r  = \sum_N b_i e^{iqr}  / \sum_N b_i

    The forward scattering per particle is (the latter only for homogeneous particles)

    .. math:: I_0=(\int_V b(r) \mathrm{d}r )^2= V_p^2(\rho_{particle}-\rho_{solvent})^2

    Here :math:`V_p` is particle volume and :math:`\rho` is the average scattering length density.

    For polymer like particles (e.g. Gaussian chain) of :math:`N` monomers with monomer partial volume
    :math:`V_{monomer}` the particle volume is :math:`V_p=N V_{monomer}`.

    The solution forward scattering :math:`c_pI_0` can be calculated from the monomer concentration as

    .. math:: c_pI_0 = c_p V_p^2(\rho_{particle}-\rho_{solvent})^2 =
                      c_{monomer} N V_{monomer}^2(\rho_{monomer}-\rho_{solvent})^2


.. collapse:: Arbitrary shaped particles

    The scattering of **arbitrary shaped particles** can be calculated by :py:func:`~.cloudscattering.cloudScattering`
    as a cloud of points representing the desired shape.

    Methods to build clouds of scatterers e.g. a cube decorated with spheres at the corners can be
    found in :ref:`Lattice` with examples. The advantage here is that there is no double counted overlap.

.. collapse:: Distributions of particles

    In the same way **distributions of particles** as e.g. clusters of particles or nanocrystals can be calculated.
    Oriented scattering of e.g. oriented nanoclusters can be calculated by
    :py:func:`~.cloudscattering.orientedCloudScattering`.

.. collapse:: Distribution of parameters

    **Distribution of parameters**

    Experimental data might be influenced by multimodal parameters (like multiple sizes)
    or by one or several parameters distributed around a mean value.
    See :ref:`Distribution of parameters`


.. collapse:: Example scattering length densities

    Some **scattering length densities** as guide to choose realistic values for SLD and solventSLD :
     - neutron scattering  unit nm\ :sup:`-2`:
        - D2O                            = 6.335e-6 A\ :sup:`-2` = 6.335e-4 nm\ :sup:`-2`
        - H2O                            =-0.560e-6 A\ :sup:`-2` =-0.560e-4 nm\ :sup:`-2`
        - protein                        |ap| 2.0e-6 A\ :sup:`-2` |ap| 2.0e-4 nm\ :sup:`-2`
        - gold                           = 4.500e-6 A\ :sup:`-2` = 4.500e-4 nm\ :sup:`-2`
        - SiO2                           = 4.185e-6 A\ :sup:`-2` = 4.185e-4 nm\ :sup:`-2`
        - protonated polyethylene        =-0.315e-6 A\ :sup:`-2` =-0.315e-4 nm\ :sup:`-2` *bulk density*
        - protonated polyethylene glycol = 0.64e-6 A\ :sup:`-2` = 0.64e-4 nm\ :sup:`-2` *bulk density*

     - Xray scattering  unit nm^-2:
        - D2O                            = 0.94e-3 nm\ :sup:`-2` = 332 e/nm\ :sup:`3`
        - H2O                            = 0.94e-3 nm\ :sup:`-2` = 333 e/nm\ :sup:`3`
        - protein                        |ap| 1.20e-3 nm\ :sup:`-2` |ap| 430 e/nm\ :sup:`3`
        - gold                           = 13.1e-3 nm\ :sup:`-2` =4662 e/nm\ :sup:`3`
        - SiO2                           = 2.25e-3 nm\ :sup:`-2` = 796 e/nm\ :sup:`3`
        - polyethylene                   = 0.85e-3 nm\ :sup:`-2` = 302 e/nm\ :sup:`3` *bulk density*
        - polyethylene glycol            = 1.1e-3 nm\ :sup:`-2` = 390 e/nm\ :sup:`3` *bulk density*

    Density SiO2 = 2.65 g/ml quartz; |ap| 2.2 g/ml quartz glass.

    Using bulk densities for polymers in solution might be wrong.
    E.g. polyethylene glycol (PEG) bulk has 390 e/nm³ but SAXS of PEG in water shows nearly matching conditions
    which corresponds to roughly 333 e/nm³ [Thiyagarajan et al.Macromolecules, Vol. 28, No. 23, (1995)]
    Reasons are a solvent dependent specific volume (dependent on temperature and molecular weight)
    and mainly hydration water density around PEG.



"""

from .polymer import *
from .bodies import *
from .composed import *
from .cloudscattering import *




import numpy as np
import ctypes
from ._cubature import cubature as _cython_cubature
from ._cubature import cubature_raw_callback as _cython_cubature_raw_callback

__all__ = ['ERROR_INDIVIDUAL', 'ERROR_PAIRED', 'ERROR_L2', 'ERROR_L1',
        'ERROR_LINF', 'cubature']

ERROR_INDIVIDUAL = 0
ERROR_PAIRED = 1
ERROR_L2 = 2
ERROR_L1 = 3
ERROR_LINF = 4

# maps (adaptive, vectorized) to appropriate function call
_call_map = {
    ('h', True): 'hcubature_v',
    ('h', False): 'hcubature',
    ('p', True): 'pcubature_v',
    ('p', False): 'pcubature',
    }

def cubature(func, ndim, fdim, xmin, xmax, args=tuple(), kwargs=dict(),
             abserr=1.e-8, relerr=1.e-8, norm=ERROR_INDIVIDUAL, maxEval=0,
             adaptive='h', vectorized=False):
    r"""Numerical-integration using the cubature method.

    Parameters
    ----------
    func : callable
        If ``vectorized=False`` the callable must have the form:
            ``f(x_array, *args, **kwargs)``

            where:

            - `x_array` in an array containing the `ndim` variables
               being integrated
            - `args` is a tuple containing any other arguments
              required by the function
            - `kwargs` is a dict containing any keyword arguments
              required by the function
            - the function must return a 1-D `np.ndarray` object
            - example 1, vector valued function::

                  fdim = 3
                  def func(x_array, *args, **kwargs):
                      # note that here ndim=2 (2 variables)
                      x, y = x_array
                      return np.array([x**2-y**2, x*y, x*y**2])

            - example 2, scalar function::

                  fdim = 1
                  def func(x_array, *args, **kwargs):
                      # note that here ndim=2 (2 variables)
                      x, y = x_array
                      return x**2-y**2

        If ``vectorized=True`` the function must have the form:
            ``f(x_array, *args, **kwargs)``

            where:

            - `x_array` has ``shape=(npt, ndim)``
            - `args` is a tuple containing any other arguments
              required by the function
            - `kwargs` is a dict containing any keyword arguments
              required by the function
            - example 1, vector valued function::

                  # function that returns a vector with 3 values
                  fdim = 3
                  def func(x_array, *args, **kwargs):
                      # note that here ndim=2 (2 variables)
                      x = x_array[:, 0]
                      y = x_array[:, 1]
                      npt = x_array.shape[0]
                      out = np.zeros((npt, fdim))
                      out[:, 0] = x**2 - y**2
                      out[:, 1] = x*y
                      out[:, 2] = x*y**2
                      return out

            - example 2, scalar function::

                  fdim = 1
                  def func(x_array, *args, **kwargs):
                      # note that here ndim=3 (3 variables)
                      x = x_array[:, 0]
                      y = x_array[:, 1]
                      z = x_array[:, 2]
                      return x**2 + y**2 + z**2

        The results from both vectorized and non-vectorized examples
        above should be the same, but the vectorized implementation
        is much faster since it will take advantage of NumPy's vectorization
        capabilities.
    ndim : integer
        Number dimensions or number of variables being integrated.
    fdim : integer
        Length of the output vector given by `func`. It should be `1` if the
        function returns a scalar.
    xmin : array-like
        A 1-D array carring the minimum integration limit for each
        variable being integrated. It must be have:
        ``xmin.shape[0]=ndim``.
    xmax : array-like
        A 1-D array carring the maximum integration limit for each
        variable being integrated. It must be have:
        ``xmax.shape[0]=ndim``.
    args : tuple or list, optional
        Contains the extra arguments required by `func`.
    kwargs : dict-like, optional
        Contains the extra keyword arguments passed to `func`.
    adaptive : string, optional
        The adaptive scheme used along the adaptive integration:

        - 'h' means 'h-adaptive', where the domain is partitioned
        - 'p' means 'p-adaptive', where the order of the integration rule is
          increased

        The 'p-adaptive' scheme is often better for smoth functions in
        low dimensions.
    abserr : double, optional
        Integration stops when estimated absolute error is below this threshold
    relerr : double, optional
        Integration stops when estimated error in integral value is below this
        threshold
    norm : integer, optional
        Specifies the norm that is used to measure the error and
        determine convergence properties (irrelevant for
        single-valued functions).
        The `norm` argument takes one of the values:

        - ERROR_INDIVIDUAL: convergence is achieved only when each
            integrand individually satisfies the requested error
            tolerances;
        - ERROR_PAIRED: like ERROR_INDIVIDUAL, except that the
            integrands are grouped into consecutive pairs, with the error
            tolerance applied in a L2 sense to each pair. This option is
            mainly useful for integrating vectors of complex numbers,
            where each consecutive pair of real integrans is the real
            and imaginary parts of a single complex integrand, and you
            only care about the error in the complex plane rather than
            the error in the real and imaginary parts separately;
        - ERROR_L2
        - ERROR_L1
        - ERROR_LINF
            the absolute error is measured as |e| and the relative error
            as |err|/|val|, where |...| is the L1, L2, or L-infinity
            norm, respectively.  (|x| in the L1 norm is the sum of the
            absolute values of the components, in the L2 norm is the
            root mean square of the components, and in the L-infinity
            norm is the maximum absolute value of the components).

    maxEval : integer, optional
        The maximum number of function evaluations.
    vectorized : boolean, optional
        If ``vectorized=True`` the integration points are passed to the
        integrand function as an array of points, allowing parallel
        evaluation of different points.

    Returns
    -------
    val : numpy.ndarray
        The 1-D array of length ``fdim`` with the computed integral values
    err : numpy.ndarray
        The 1-D array of length ``fdim`` with the estimated errors. For
        smooth functions this estimate is usually conservative (see the
        results from the ``test_cubature.py`` script.

    Notes
    -----
    * The supplied function must return a 1-D ``np.ndarray`` object,
      even for single-valued functions

    References
    ----------
    .. [1] `Cubature (Multi-dimensional integration)
           <http://ab-initio.mit.edu/wiki/index.php/Cubature>`_.

    Examples
    --------

    >>> # Volume of a sphere:
    >>> import numpy as np
    >>> from cubature import cubature

    >>> def integrand_sphere(x_array, *args):
    >>>     r, theta, phi = x_array
    >>>     return np.array([r**2*sin(phi)])

    >>> ndim = 3
    >>> fdim = 1
    >>> radius = 1.
    >>> xmin = np.array([0, 0, 0])
    >>> xmax = np.array([radius, 2*pi, pi])
    >>> val, err = cubature(integrand_sphere, ndim, fdim, xmin, xmax)

    """
    # checking xmin and xmax
    xmin = np.asarray(xmin)
    xmax = np.asarray(xmax)
    try:
        assert xmin.shape[0] == ndim
    except:
        raise ValueError('xmin.shape[0] is not equal ndim')
    try:
        assert xmax.shape[0] == ndim
    except:
        raise ValueError('xmax.shape[0] is not equal ndim')

    use_raw_callback = isinstance(func, ctypes._CFuncPtr)

    # checking fdim
    if not use_raw_callback:
        if not vectorized:
            out = func(np.ones(ndim)*xmin, *args, **kwargs)
            try:
                if isinstance(out, float) or isinstance(out, int):
                    out = np.array([out])
                assert out.shape[0] == fdim
            except:
                raise ValueError(
                    'Length of func ouptut vector is different than fdim')
        else:
            out = func(np.ones((7, ndim))*xmin, *args, **kwargs)
            if fdim > 1:
                try:
                    assert out.shape[0] == 7
                    assert out.shape[1] == fdim
                except:
                    raise ValueError(
                        'Output vector does not have shape=(:, fdim)')
            else:
                try:
                    assert out.ndim == 1
                    assert out.shape[0] == 7
                except:
                    raise ValueError(
                        'Output vector does not return a valid array')

    method = _call_map.get((adaptive, vectorized), None)
    if method is None:
        s = 'unknown combination of adaptive (`{!r}`) and vectorized (`{!r}`).'
        s = s.format(adaptive, vectorized)
        raise ValueError(s)
    else:
        if use_raw_callback:
            val, err = _cython_cubature_raw_callback(func, ndim, fdim, xmin, xmax,
                    method, abserr, relerr, norm, maxEval, args=args, kwargs=kwargs)
        else:
            val, err = _cython_cubature(func, ndim, fdim, xmin, xmax, method, abserr,
                    relerr, norm, maxEval, args=args, kwargs=kwargs)

    return val, err




# testfile adopted from original source but converted to use unittest
# use by
# python3 -m unittest jscatter.libs.cubature.test_cubature

from . import _test_integrands as ti
from . import cubature
import math
import numpy as np
import unittest


class cubatureTest(unittest.TestCase):
    # test that the Genz oscillatory exact formula actually agrees at an
    # even and odd dimension
    def test_genz_oscillatory_exact_d3(self):
        exact = 4 * np.cos(1056 + np.pi/7)*np.sin(12)*np.sin(306)*np.sin(738)/6273
        calculated = ti.genz_oscillatory_exact(12, np.array([2, 123, 51]), 1/14)
        assert np.allclose([exact], [calculated])

    def test_genz_oscillatory_exact_d4(self):
        exact = np.sin(29988 - 7*np.pi/34) * \
                np.sin(51)*np.sin(2601/2)*np.sin(6273/2)*np.sin(25500) / \
                784125
        calculated = ti.genz_oscillatory_exact(51, np.array([2, 123, 51, 1000]),
                11/17)
        assert np.allclose([exact], [calculated])

    def test_genz_oscillatory_d2(self):
        x = np.array([17, 201], dtype=float)
        a = np.array([41, 1/11], dtype=float)
        u = 1/51

        exact = np.cos(7868/11 + 2*np.pi/51)
        calculated = ti.genz_oscillatory(x, a, u)
        assert np.allclose([exact], [calculated])

    def test_genz_oscillatory_c_d2(self):
        x = np.array([17, 201], dtype=float)
        a = np.array([41, 1/11], dtype=float)
        u = 1/51

        exact = np.cos(7868/11 + 2*np.pi/51)
        calculated = ti.genz_oscillatory(x, a, u)
        calculated_2 = ti.genz_oscillatory_c(x, a, u)

        assert np.allclose([exact], [calculated_2])
        assert np.allclose([calculated], [calculated_2])

    def test_hcubature_genz_oscillatory_d2(self):
        u = 2*np.pi*15/609
        a = np.array([15.51, 2], dtype=float)
        n = 3

        xmin = np.zeros((2,))
        xmax = np.ones((2,)) * n

        exact = ti.genz_oscillatory_exact(n, a, u)

        # check that integrand is callable
        ti.genz_oscillatory(np.ones((2,), dtype=float), a, u)

        val, err = cubature(ti.genz_oscillatory, 2, 1, xmin, xmax, args=(a, u),
                adaptive='h')
        assert np.allclose(exact, val)

    def test_cubature_zero_exact(self):
        d = 1
        xmin = np.zeros((d,), dtype=float)
        xmax = np.ones((d,), dtype=float)*np.pi/2

        assert np.allclose(0., ti.cubature_zero_exact(xmin))
        assert np.allclose(1., ti.cubature_zero_exact(xmax))

        d = 3
        xmin = np.zeros((d,), dtype=float)
        assert np.allclose(0., ti.cubature_zero_exact(xmin))
        assert np.allclose(1., ti.cubature_zero_exact(xmax))

        d = 4
        xmin = np.zeros((d,), dtype=float)
        assert np.allclose(0., ti.cubature_zero_exact(xmin))
        assert np.allclose(1., ti.cubature_zero_exact(xmax))

    def test_cubature_zero(self):
        d = 5
        k = 2/np.sqrt(np.pi)
        x = np.ones((d,), dtype=float)
        expected = k**d
        assert np.allclose(ti.cubature_one(x), expected)

    def test_hcubature_cubature_zero(self):
        xmin = np.zeros((4,))
        xmax = np.array([12, 4, 0.25, 1], dtype=float)

        # check that it is possible to run
        ti.cubature_zero(xmax)

        exact = ti.cubature_zero_exact(xmax)
        val, err = cubature(ti.cubature_zero, xmax.shape[0], 1, xmin, xmax)

        assert np.allclose([exact], [val])

    def test_hcubature_cubature_one(self):
        d = 3
        xmax = np.ones((d,), dtype=float)
        xmin = np.zeros_like(xmax)

        ti.cubature_one(xmin)
        exact = ti.cubature_one_exact(xmax)
        val, err = cubature(ti.cubature_one, d, 1, xmin, xmax)
        assert np.allclose([exact], [val])

    def test_cubature_two_exact(self):
        radius = 1.

        d = 1
        exact = 2
        val = ti.cubature_two_exact(d, radius)
        assert np.allclose([exact], [val])

        d = 2
        exact = np.pi
        val = ti.cubature_two_exact(d, radius)
        assert np.allclose([exact], [val])

        d = 3
        exact = 4/3*np.pi
        val = ti.cubature_two_exact(d, radius)
        assert np.allclose([exact], [val])

    def test_hcubature_cubature_two(self):
        radius = 0.68244456511919859846
        d = 2
        xmin = -np.ones((d,))
        xmax = np.ones((d,))

        exact = ti.cubature_two_exact(d, radius)

        val, err = cubature(ti.cubature_two, d, 1, xmin, xmax, args=(radius,),
                abserr=1e-4, relerr=1e-4, maxEval=1000000)
        true_error = np.abs(val - exact)
        assert true_error < 1e-4

    def test_cubature_three_exact(self):
        d = 8
        xmin = -np.ones((d,))
        xmax = np.ones((d,))

        val = ti.cubature_three_exact(xmin, xmax)
        assert np.abs(val) < 1e-14

        xmin = np.zeros((d,))
        xmax = np.ones((d,))

        val = ti.cubature_three_exact(xmin, xmax)
        assert np.abs(val - 1) < 1e-14

    def test_hcubature_cubature_three(self):
        d = 8
        xmin = np.zeros((d,))
        xmax = np.ones((d,))

        exact = ti.cubature_three_exact(xmin, xmax)

        val, err = cubature(ti.cubature_three, d, 1, xmin, xmax)

        assert np.allclose([exact], [val])

    def test_genz_gaussian_exact(self):
        u = np.array([1, 21, 2], dtype=float)
        a = np.array([1/10, 1/100, 1/500], dtype=float)
        val = ti.genz_gaussian_exact(u, a)
        exact = 62500*pow(np.pi, 3/2)*math.erf(1/10)*(math.erf(1/250) -
                math.erf(1/500))*(math.erf(21/100) - math.erf(1/5))

        assert np.allclose([val], [exact])

    def test_hcubature_genz_gaussian(self):
        u = np.array([1, 21, 2], dtype=float)
        a = np.array([1/10, 1/100, 1/500], dtype=float)

        xmin = np.zeros_like(u)
        xmax = np.ones_like(u)

        exact = ti.genz_gaussian_exact(u, a)

        val, err = cubature(ti.genz_gaussian, u.shape[0], 1, xmin, xmax, args=(u, a),
                adaptive='h')

        assert np.allclose([exact], [val])

    def _test_hcubature_cubature_five(self):
        d = 3
        xmin = -np.ones((d,))
        xmax = np.ones((d,))

        exact = 1.

        val, err = cubature(ti.cubature_five, d, 1, xmin, xmax)

        assert np.allclose([exact], [val])

    def test_hcubature_cubature_six(self):
        d = 3
        xmin = np.zeros((d,))
        xmax = np.ones((d,))

        exact = 1.

        val, err = cubature(ti.cubature_six, d, 1, xmin, xmax)

        assert np.allclose([exact], [val])

    def _test_hcubature_cubature_seven(self):
        d = 4
        xmin = np.zeros((d,))
        xmax = np.ones((d,))

        exact = 1.

        val, err = cubature(ti.cubature_seven, d, 1, xmin, xmax)

        assert np.allclose([exact], [val])


def suite():
    loader = unittest.TestLoader()
    s = unittest.TestSuite()
    s.addTest(loader.loadTestsFromTestCase(cubatureTest))

    return s


def doTest(verbosity=1):
    runner = unittest.TextTestRunner(verbosity=verbosity)
    a = runner.run(suite())
    return a


if __name__ == '__main__':
    unittest.main(defaultTest=doTest)




"""
A minimal example how to fit Iqt from NSE measurements on polyelectrolytes using finiteZimm
to demonstarte how to include H(Q) and S(Q) for Dcm.

For details on the physics see

Interchain Hydrodynamic Interaction and Internal Friction of Polyelectrolytes
Buvalaia, et al ACS Macro Letters 12(9):1218    DOI: 10.1021/acsmacrolett.3c00409

"""
import jscatter as js
import numpy as np

# the structure factor type and parameters should be determined by fitting of SAXS/SANS data
# only for simplicity we use the PercusYevick, molarity is determined by experimental conditions of NSE
# here the experimental R of the PercusYevick is determined from experimental data

q0 = np.r_[0.01:2:100j]


def collectivefiniteZIMM(t, q, Dt=None, mu=0.5, Rh=4, R=4,tint=0, mol=0.0003):
    # Hq can be determined integrating over Sq and using Rh as a fit parameter
    # We assume here a polymer with different hydrodynamic interaction compared to a sphere
    # internally SF is calculated and saved in _sf
    SF = js.sf.PercusYevick
    HqSq = js.sf.hydrodynamicFunct(q0, Rh=Rh, molarity=mol,
                                 structureFactor=SF,
                                 structureFactorArgs={'R': R}, )
    HqSq.Y =  HqSq.Y / HqSq._sf  # this makes the needed H(Q)/S(Q)
    hqsq = HqSq[:2]  # only first columns

    # Dt is determined inside finiteZimm , but can be changed
    Iqt = js.dynamic.finiteZimm(t, q, NN=212, pmax=25, l=0.24, Dcm=Dt, tintern=tint, Temp=273 + 60, mu=mu,
                                viscosity=1, Dcmfkt = hqsq)
    Iqt.setColumnIndex(iey=None)
    return Iqt


# simulate data, typical NSE times IN15@ILL
tlist = js.loglist(0.1,300,30)

p = js.grace(2,1)
p.multi(1,2)
p[0].xaxis(label='t / ns',scale='log')
p[1].xaxis(label='t / ns',scale='log')
p[0].yaxis(label='I(Q,t)/I(Q,0)')

# lowest Q corresponds to DLS as synthetic dataset added to Iqt from NSE
sim = js.dL()
for q in np.r_[0.026,0.25,0.4,0.8,1.2,1.4,1.8]:  # in 1/nm
    sim.append(collectivefiniteZIMM(t=tlist, q=q, mol=0.0003, R=8, Rh=4, mu=0.6))
p[0].plot(sim,sy=0,li=[1,3,1])
p[1].plot(sim,sy=0,li=[1,3,1])

sim2 = js.dL()
for q in np.r_[0.026,0.25,0.4,0.6,0.8,1.2,1.4]:  # in 1/nm
    sim2.append(collectivefiniteZIMM(t=tlist, q=q, mol=0.0003, R=7, Rh=4, mu=0.6))
p[0].plot(sim2,sy=0,li=[3,3,2])

sim3 = js.dL()
for q in np.r_[0.026,0.25,0.4,0.6,0.8,1.2,1.4]:  # in 1/nm
    sim3.append(collectivefiniteZIMM(t=tlist, q=q, mol=0.0003, R=8, Rh=6, mu=0.6))
p[1].plot(sim3,sy=0,li=[3,3,4])

p[0].text('DLS',x=300,y=0.8)
p[1].text('DLS',x=300,y=0.8)
p[0].title('Interaction influences low Q, specifically DLS')
p[1].title('Rh influences all Q as self diffusion changes')

# p.save('collectiveZimm.png', size=(2, 1), dpi=150)

if 0:
    # A fit to exp. NSE data might be done like this (using sim as measured data, no errors)
    # fixpar are determined from other experiments
    sim.fit(model=collectivefiniteZIMM,
                freepar={'Rh':4,'mu':0.6,'tint':1},
                fixpar={'R':6, 'mol':0.0003 },  # from exp. conditions
                mapNames={'t':'X'})





import numpy as np
import jscatter as js

# data to fit
# We can fit all as it is faster than bayes
i5 = js.dL(js.examples.datapath+'/iqt_1hho.dat')

# model
def diffusion(A, D, t, elastic, wavevector=0):
    return A*np.exp(-wavevector**2*D*t) + elastic

# Define regularisation function that describes knowledge about the data, parameters or how parameters are distributed.
Amean, Asig = 1.0 , 0.01
Dmean, Dsig = 0.09 , 0.02
def ln_priorGAUSS(A, D):
    # assuming a normalised Gaussian distribution around the mean of A and D with width sig (Gaussian prior)
    # the log of the Gaussian is ~ lambda*w**2 = 0.5/Asig**2 * (a-Amean)**2,
    # lambda*w**2 should be always positive without normalisation.
    # the parameters are arrays for all elements of the dataList or a float for common parameters
    # the 't' is not included as it describes the .X values, 'elastic' is not used (no knowledge about it).
    lp = 0.5 * np.sum((A-Amean)**2/Asig**2)
    lp += 0.5 * np.sum((D-Dmean)**2/Dsig**2)
    return lp

# as example => Laplace prior.
# The Laplace prior has fatter tails than the Gaussian, and is also more concentrated around zero.
def ln_priorLAPLACE(A, D):
    # while Gauss uses L2 norm , here we use L1 norm =>|A-Amean|
    lp =  np.sum(np.abs(A-Amean)/(2**0.5*Asig))
    lp += np.sum(np.abs(D-Dmean)/2**0.5*Dsig)
    return lp

# as example => Gaussian around the average only for A and usage of current modelValues.
# if ln_prior has keyword 'modelValues' these are added from the current calculation
def ln_priorMEAN(modelValues, A):
    # for a free A we want that these are within some sigma with unknown mean.
    lp =  np.sum(np.abs(A-np.mean(A))**2) / Asig**2
    # this is a stupid example as demonstration
    # e.g in SAXS it might be I0 or Rg that determine a penalty
    # that is not in the parameters but in the formfactor calculation
    lp += 0 if modelValues.Y[:5,0].mean > 0.9 else 10

    return lp


i5.setlimit(D=[0.05, 1], A=[0.5, 1.5])

# do the fit with regression constrain in  ''ln_prior''.
i5.fit(model=diffusion, freepar={'D': [0.2], 'A': 0.98}, fixpar={'elastic': 0.0},
      mapNames={'t': 'X', 'wavevector': 'q'}, condition=lambda a: a.X<90,
      method='lm', ln_prior=ln_priorGAUSS)

i5.showlastErrPlot()

p=js.grace()
p.plot(i5.q,i5.D,i5.D_err,le='To strong regularisation')

# The result is not good as the constraint on D is to much constraining the diffusion coefficient.
# Therefore the errors are large and the values to close to 0.09.
# better is simple 'lm' without constrain in this case. This was wrong knowledge!

i5.fit(model=diffusion, freepar={'D': [0.2], 'A': 0.98}, fixpar={'elastic': 0.0},
       mapNames={'t': 'X', 'wavevector': 'q'}, condition=lambda a: a.X<90,
       method='lm',output=False)
p.plot(i5.q,i5.D,i5.D_err,le='LevenbergMarquardt fit')
p.legend(x=0.1,y=0.13)
p.yaxis(min=0.05,max=0.15)






# coding=utf-8

# Crystal structure factors in 2D

# Comparison of different domain sizes dependent on direction of scattering ::

import jscatter as js
import numpy as np

# make xy grid in q space
R = 8  # maximum
N = 50  # number of points
qxy = np.mgrid[-R:R:N * 1j, -R:R:N * 1j].reshape(2, -1).T
# add z=0 component
qxyz = np.c_[qxy, np.zeros(N ** 2)]
# create sc lattice which includes reciprocal lattice vectors and methods to get peak positions
sclattice = js.lattice.scLattice(2.1, 5)
sclattice.rotatehkl2Vector([1, 0, 0], [0, 0, 1])
# define crystal size in directions
ds = [[20, 1, 0, 0], [5, 0, 1, 0], [5, 0, 0, 1]]
# We orient to 100 direction perpendicular to center of qxy plane
ffs = js.sf.orientedLatticeStructureFactor(qxyz, sclattice, domainsize=ds, rmsd=0.1, hklmax=2)
fig = js.mpl.surface(qxyz[:, 0], qxyz[:, 1], ffs[3].array)
fig.axes[0].set_title('symmetric peaks: thinner direction perpendicular to scattering plane')
fig.show()
# We orient to 010 direction perpendicular to center of qxy plane
sclattice.rotatehkl2Vector([0, 1, 0], [0, 0, 1])
ffs = js.sf.orientedLatticeStructureFactor(qxyz, sclattice, domainsize=ds, rmsd=0.1, hklmax=2)
fig2 = js.mpl.surface(qxyz[:, 0], qxyz[:, 1], ffs[3].array)
fig2.axes[0].set_title('asymmetric peaks: thin direction is parallel to scattering plane')
fig2.show()

# rhombic lattice simple and body centered

import jscatter as js
import numpy as np

# make xy grid in q space
R = 8  # maximum
N = 50  # number of points
qxy = np.mgrid[-R:R:N * 1j, -R:R:N * 1j].reshape(2, -1).T
# add z=0 component
qxyz = np.c_[qxy, np.zeros(N ** 2)]
# create rhombic  bc lattice which includes reciprocal lattice vectors and methods to get peak positions
rblattice = js.lattice.rhombicLattice([[2, 0, 0], [0, 3, 0], [0, 0, 1]], size=[5, 5, 5],
                                      unitCellAtoms=[[0, 0, 0], [0.5, 0.5, 0.5]])
# We orient to 100 direction perpendicular to xy plane
rblattice.rotatehkl2Vector([1, 0, 0], [0, 0, 1])

# define crystal size in directions
ds = [[20, 1, 0, 0], [5, 0, 1, 0], [5, 0, 0, 1]]
ffs = js.sf.orientedLatticeStructureFactor(qxyz, rblattice, domainsize=ds, rmsd=0.1, hklmax=2)
fig = js.mpl.surface(ffs.X, ffs.Z, ffs[3].array)
fig.axes[0].set_title('rhombic body centered lattice')
fig.show()
# same without body centered atom
tlattice = js.lattice.rhombicLattice([[2, 0, 0], [0, 3, 0], [0, 0, 1]], size=[5, 5, 5])
tlattice.rotatehkl2Vector([1, 0, 0], [0, 0, 1])
ffs = js.sf.orientedLatticeStructureFactor(qxyz, tlattice, domainsize=ds, rmsd=0.1, hklmax=2)
fig2 = js.mpl.surface(ffs.X, ffs.Z, ffs[3].array)
fig2.axes[0].set_title('rhombic lattice')
fig2.show()

# Rotation of 10 degrees along [1,1,1] axis. It looks spiky because of low number of points in xy plane ::

import jscatter as js
import numpy as np

# make xy grid in q space
R = 8  # maximum
N = 800  # number of points
qxy = np.mgrid[-R:R:N * 1j, -R:R:N * 1j].reshape(2, -1).T
# add z=0 component
qxyz = np.c_[qxy, np.zeros(N ** 2)]  # as position vectors
# create sc lattice which includes reciprocal lattice vectors and methods to get peak positions
sclattice = js.lattice.scLattice(2.1, 5)
# We orient to 111 direction perpendicular to xy plane
sclattice.rotatehkl2Vector([1, 1, 1], [0, 0, 1])
# this needs crystal rotation by 15 degrees to be aligned to xy plane after rotation to 111 direction
# The crystals rotates by 10 degrees around 111 to broaden peaks.
ds = 15
fpi = np.pi / 180.
ffs = js.sf.orientedLatticeStructureFactor(qxyz, sclattice, rotation=[1, 1, 1, 10 * fpi],
                                           domainsize=ds, rmsd=0.1, hklmax=2, nGauss=23)
fig = js.mpl.surface(ffs.X, ffs.Z, ffs[3].array)
fig.axes[0].set_title('10 degree rotation around 111 direction')
fig.show()



import jscatter as js
import numpy as np

x = js.loglist(0.02, 100, 100)  # np.r_[0.02:0.3:0.02,0.3:3:0.2,3:40:1]
qmax = 2
step = 0.15
NN = 124  # number of beads
qq = np.r_[0.1:qmax:step]

pp = js.grace(0.75, 1)
pp.multi(2, 2, hgap=0)
p0 = pp[2]
p1 = pp[0]
p2 = pp[3]
p3 = pp[1]

# Zimm dynamics with hydrodynamic interaction
zz1 = js.dynamic.finiteZimm(x, qq, 124, 7, l=0.38, tintern=0., mu=0.5, Temp=273 + 60, viscosity=.565)

# Rouse dynamics without hydrodynamic interaction
rr1 = js.dynamic.finiteRouse(x, qq, 124, 7, l=0.38, Dcm=zz1.Dcm[0], tintern=0., Temp=273 + 60)

p1.title(r'Rouse dynamics (NO hydrodynamics)', size=1)
for i, z in enumerate(rr1, 1):
    p1.plot(z.X, z.Y, line=[1, 1, i], symbol=0, legend='q=%g' % z.q)
    p1.plot(z.X, z[2], line=[3, 2, i], symbol=0, legend='q=%g diff' % z.q)

for i, mc in enumerate(np.array(rr1.modecontribution).T, 1):
    p0.plot(rr1.q, mc, li=[1, 2, i], sy=0, legend='mode %i' % i)

p1.yaxis(min=0.0, max=1.0, scale='n', label='I(q,t)/I(q,0)')
p1.xaxis(min=0.02, max=max(x), scale='l')
p0.yaxis(min=0., max=1.1, scale='n', label='mode contribution factors')
p0.xaxis(min=0.0, max=qmax, scale='n', label=r'q / nm\S-1')
# p1.legend(x=3., y=0.6, charsize=0.5)
p0.legend(x=0.1, y=0.8, charsize=0.5)
p1.subtitle(r'\xt\f{}\sRouse\N=%.3g ns, \xt\f{}\sintern\N=%.3g ns' % (rr1.trouse[0], rr1.tintern[0]))

p3.title(r'Zimm dynamics (incl hydrodynamics)', size=1)
for i, z in enumerate(zz1, 1):
    p3.plot(z.X, z.Y, line=[1, 1, i], symbol=0, legend='q=%g' % z.q)
    p3.plot(z.X, z[2], line=[3, 2, i], symbol=0, legend='q=%g diff' % z.q)

for i, mc in enumerate(np.array(zz1.modecontribution).T, 1):
    p2.plot(zz1.q, mc, li=[1, 2, i], sy=0, legend='mode %i' % i)

p3.yaxis(min=0., max=1.0, scale='n', ticklabel=0)
p3.xaxis(min=0.02, max=max(x), scale='l')
p2.yaxis(min=0.0, max=1.1, scale='n', ticklabel=0)
p2.xaxis(min=0.0, max=qmax, scale='n', label=r'q / nm\S-1')
p3.legend(x=99.3, y=1.0, charsize=0.5)
p2.legend(x=0.1, y=0.8, charsize=0.5)
p3.subtitle(r'\xt\f{}\sZimm\N=%.3g ns , \xt\f{}\sintern\N=%.3g ns' % (z.tzimm, z.tintern))
pp.save('ZimmDynamics.png')



# -*- coding: utf-8 -*-
# written by Ralf Biehl at the Forschungszentrum Jülich ,
# Jülich Center for Neutron Science 1 and Institute of Complex Systems 1
#    Jscatter is a program to read, analyse and plot data
#    Copyright (C) 2015-2024  Ralf Biehl
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

r"""
Models describing dynamic processes mainly for inelastic neutron scattering.

- Models in the time domain have a parameter t for time. -> intermediate scattering function :math:`I(t,q)`
- Models in the frequency domain have a parameter w for frequency and _w appended. ->
  dynamic structure factor :math:`S(w,q)`

Models in time domain can be transformed to frequency domain by :py:func:`~.fft.time2frequencyFF`
implementing the Fourier transform :math:`S(w,q)=F(I(t,q))`.

In time domain the combination of processes :math:`I_i(t,q)` is done by multiplication,
including instrument resolution :math:`R(t,q)`:

:math:`I(t,q)=I_1(t,q)I_2(t,q)R(t,q)`.
::

 # multiplying and creating new dataArray
 I(t,q) = js.dA( np.c[t, I1(t,q,..).Y*I2(t,q,..).Y*R(t,q,..).Y ].T)

In frequency domain it is a convolution, including the instrument resolution.

:math:`S(w,q) = S_1(w,q) \otimes S_2(w,q) \otimes R(w,q)`.
::

 conv=js.formel.convolve
 S(w,q)=conv(conv(S1(w,q,..),S2(w,q,..)),res(w,q,..),normB=True)      # normB normalizes resolution

res(w,q,..) might be a measurement or the result of a fit to a measurement.
A fitted measurement assumes a continous extrapolation at the edges and a smoothing, while
a measurement is cut at the edges to zero and contains noise.

.. collapse:: Time to frequency FFT

    FFT from time domain by :py:func:`~.fft.time2frequencyFF` may include the resolution where it acts like a
    window function to reduce spectral leakage with vanishing values at :math:`t_{max}`.
    If not used :math:`t_{max}` needs to be large (see tfactor) to reduce spectral leakage.

.. collapse:: shiftAndBinning

    The last step is to shift the model spectrum to the symmetry point of the instrument
    as found in the resolution measurement and optional binning over frequency channels.
    Both is done by :py:func:`~.fft.shiftAndBinning`.

.. collapse:: Usage example

    **Example**

    Let us describe the diffusion of a particle inside a diffusing invisible sphere
    by mixing time domain and frequency domain.
    ::

     resolutionparameter={'s0':5,'m0':0,'a0':1,'bgr':0.00}
     w=np.r_[-100:100:0.5]
     resolution=js.dynamic.resolution_w(w,**resolutionparameter)
     # model
     def diffindiffSphere(w,q,R,Dp,Ds,w0,bgr):
         # time domain with transform to frequency domain
         diff_w=js.dynamic.time2frequencyFF(js.dynamic.simpleDiffusion,resolution,q=q,D=Ds)
         # last convolution in frequency domain, resolution is already included in time domain.
         Sx=js.formel.convolve(js.dynamic.diffusionInSphere_w(w=w,q=q,D=Dp,R=R),diff_w)
         Sxsb=js.dynamic.shiftAndBinning(Sx,w=w,w0=w0)
         Sxsb.Y+=bgr       # add background
         return Sxsb
     #
     Iqw=diffindiffSphere(w=w,q=5.5,R=0.5,Dp=1,Ds=0.035,w0=1,bgr=1e-4)


    For more complex systems e.g a number x of diffusing hydrogens and a number y of fixed oxygen and carbons
    different scattering length and changing fraction of contributing atoms (with scattering length)
    has to be included just by adding a corresponding factor. This factor can e.g. be determined from an atomic structure.

    Accordingly, if desired, the mixture of coherent and incoherent scattering needs to be accounted for
    by corresponding scattering length.
    This additionally is dependent on the used instrument e.g. for spin echo only 1/3 of the incoherent scattering
    contributes to the signal.
    An example model for protein dynamics is given in :ref:`Protein incoherent scattering in frequency domain`.

A comparison of different dynamic models in frequency domain is given in examples.
:ref:`A comparison of different dynamic models in frequency domain`.



"""


from .timedomain import *
from .frequencydomain import *
from .fft import *




# -*- coding: utf-8 -*-
#  written by Ralf Biehl at the Forschungszentrum Juelich
# Juelich Center for Neutron Science 1 and Institute of Complex Systems 1
# ra.biehl@fz-juelich.de
#    jscatter is a program to read, analyse and plot data
#    Copyright (C) 2015-2024  Ralf Biehl
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

def _w2f(word):
    """
    Converts strings if possible to float.
    """
    try:
        return float(word)
    except ValueError:
        return word


def _dictfromlist(lines):
    """
    """
    d = {}
    lines = lines.splitlines()
    for line in lines:
        if line.startswith("#"):
            continue
        words = line.split()
        if len(words) == 0 :
            continue
        d[_w2f(words[0])] = [_w2f(w) for w in words[1:] ]
    return d


def getvdWdict(lines):
    """
    return most probable wan der Waals radius in nm

    default is 0 if data are missing  for Z>92

    use as (see formel):

     vdwradii = getvdWdict(tableofvdWRadii)
     vdwradii.update({'D': vdwradii['H']})

    """
    d = {}
    lines = lines.splitlines()
    for line in lines:
        if line.startswith("#"): continue
        words = line.split()
        if len(words) == 0 :    continue
        try:
            # vdW
            d[words[1]] = float(words[4])/1000
        except ValueError:
            try:
                # calculated
                d[words[1]] = float(words[3])/1000
            except ValueError:
                try:
                    # empirical
                    d[words[1]] = float(words[2])/1000
                except ValueError:
                    try:
                        d[words[1]] = float(words[5])/1000
                    except ValueError:
                        d[words[1]] = 0
    return d


#: van der Waals radii from
#: https://en.wikipedia.org/wiki/Atomic_radii_of_the_elements_(data_page)#Van_der_Waals_radius
tableofvdWRadii="""
# From
# https://en.wikipedia.org/wiki/Atomic_radii_of_the_elements_(data_page)#Van_der_Waals_radius
#
# References:
# Atomic radius (empirical) 
#     J.C. Slater (1964). "Atomic Radii in Crystals". 
#     J. Chem. Phys. 41: 3199. Bibcode:1964JChPh..41.3199S. doi:10.1063/1.1725697
# Atomic radius (calculated) 
#     E. Clementi; D.L.Raimondi; W.P. Reinhardt (1967). 
#         "Atomic Screening Constants from SCF Functions. II. Atoms with 37 to 86 Electrons".
#         J. Chem. Phys. 47: 1300. Bibcode:1967JChPh..47.1300C. doi:10.1063/1.1712084.
# Van der Waals radius
#    A. Bondi (1964). "van der Waals Volumes and Radii". J. Phys. Chem. 68: 441. doi:10.1021/j100785a001.
#    M. Mantina; A.C. Chamberlin; R. Valero; C.J. Cramer; D.G. Truhlar (2009). 
#           "Consistent van der Waals Radii for the Whole Main Group". 
#           J. Phys. Chem. A. 113 (19): 5806–12. Bibcode:2009JPCA..113.5806M. doi:10.1021/jp8111556. PMID 19382751.
# Covalent radii (single bond)
# The above values are based on
#    R.T. Sanderson (1962). Chemical Periodicity. New York, USA: Reinhold.
#    L.E. Sutton, ed. (1965). "Supplement 1956–1959, Special publication No. 18". 
#         Table of interatomic distances and configuration in molecules and ions. London, UK: Chemical Society.
#    J.E. Huheey; E.A. Keiter & R.L. Keiter (1993). Inorganic Chemistry : 
#         Principles of Structure and Reactivity (4th ed.). New York, USA: HarperCollins. ISBN 0-06-042995-X.
#    W.W. Porterfield (1984). Inorganic chemistry, a unified approach. 
#         Reading Massachusetts, USA: Addison Wesley Publishing Co. ISBN 0-201-05660-7.
#    A.M. James & M.P. Lord (1992). Macmillan's Chemical and Physical Data. 
#         MacMillan. ISBN 0-333-51167-0.
#
# Jscatter uses the van der Waals value. If missing the calculated, then empirical, then covalent value is used.
# All measurements given are in picometers (pm)
#
# For H we use the value 109 pm (instead of 120pm) from Rowland  J Phys. chem. 100, 7384 (1996)

# atomic number | symbol | empirical | calculated | van der Waals | covalent

1.       H   25.0   53.0  109.0   38.0
2.      He  120.0   31.0  140.0   32.0
3.      Li  145.0  167.0  182.0  134.0
4.      Be  105.0  112.0  153.0   90.0
5.       B   85.0   87.0  192.0   82.0
6.       C   70.0   67.0  170.0   77.0
7.       N   65.0   56.0  155.0   75.0
8.       O   60.0   48.0  152.0   73.0
9.       F   50.0   42.0  147.0   71.0
10.     Ne  160.0   38.0  154.0   69.0
11.     Na  180.0  190.0  227.0  154.0
12.     Mg  150.0  145.0  173.0  130.0
13.     Al  125.0  118.0  184.0  118.0
14.     Si  110.0  111.0  210.0  111.0
15.      P  100.0   98.0  180.0  106.0
16.      S  100.0   88.0  180.0  102.0
17.     Cl  100.0   79.0  175.0   99.0
18.     Ar   71.0   71.0  188.0   97.0
19.      K  220.0  243.0  275.0  196.0
20.     Ca  180.0  194.0  231.0  174.0
21.     Sc  160.0  184.0  211.0  144.0
22.     Ti  140.0  176.0 nodata  136.0
23.      V  135.0  171.0 nodata  125.0
24.     Cr  140.0  166.0 nodata  127.0
25.     Mn  140.0  161.0 nodata  139.0
26.     Fe  140.0  156.0 nodata  125.0
27.     Co  135.0  152.0 nodata  126.0
28.     Ni  135.0  149.0  163.0  121.0
29.     Cu  135.0  145.0  140.0  138.0
30.     Zn  135.0  142.0  139.0  131.0
31.     Ga  130.0  136.0  187.0  126.0
32.     Ge  125.0  125.0  211.0  122.0
33.     As  115.0  114.0  185.0  119.0
34.     Se  115.0  103.0  190.0  116.0
35.     Br  115.0   94.0  185.0  114.0
36.     Kr nodata   88.0  202.0  110.0
37.     Rb  235.0  265.0  303.0  211.0
38.     Sr  200.0  219.0  249.0  192.0
39.      Y  180.0  212.0 nodata  162.0
40.     Zr  155.0  206.0 nodata  148.0
41.     Nb  145.0  198.0 nodata  137.0
42.     Mo  145.0  190.0 nodata  145.0
43.     Tc  135.0  183.0 nodata  156.0
44.     Ru  130.0  178.0 nodata  126.0
45.     Rh  135.0  173.0 nodata  135.0
46.     Pd  140.0  169.0  163.0  131.0
47.     Ag  160.0  165.0  172.0  153.0
48.     Cd  155.0  161.0  158.0  148.0
49.     In  155.0  156.0  193.0  144.0
50.     Sn  145.0  145.0  217.0  141.0
51.     Sb  145.0  133.0  206.0  138.0
52.     Te  140.0  123.0  206.0  135.0
53.      I  140.0  115.0  198.0  133.0
54.     Xe nodata  108.0  216.0  130.0
55.     Cs  260.0  298.0  343.0  225.0
56.     Ba  215.0  253.0  268.0  198.0
57.     La  195.0  195.0 nodata  169.0
58.     Ce  185.0  158.0 nodata nodata
59.     Pr  185.0  247.0 nodata nodata
60.     Nd  185.0  206.0 nodata nodata
61.     Pm  185.0  205.0 nodata nodata
62.     Sm  185.0  238.0 nodata nodata
63.     Eu  185.0  231.0 nodata nodata
64.     Gd  180.0  233.0 nodata nodata
65.     Tb  175.0  225.0 nodata nodata
66.     Dy  175.0  228.0 nodata nodata
67.     Ho  175.0  226.0 nodata nodata
68.     Er  175.0  226.0 nodata nodata
69.     Tm  175.0  222.0 nodata nodata
70.     Yb  175.0  222.0 nodata nodata
71.     Lu  175.0  217.0 nodata  160.0
72.     Hf  155.0  208.0 nodata  150.0
73.     Ta  145.0  200.0 nodata  138.0
74.      W  135.0  193.0 nodata  146.0
75.     Re  135.0  188.0 nodata  159.0
76.     Os  130.0  185.0 nodata  128.0
77.     Ir  135.0  180.0 nodata  137.0
78.     Pt  135.0  177.0  175.0  128.0
79.     Au  135.0  174.0  166.0  144.0
80.     Hg  150.0  171.0  155.0  149.0
81.     Tl  190.0  156.0  196.0  148.0
82.     Pb  180.0  154.0  202.0  147.0
83.     Bi  160.0  143.0  207.0  146.0
84.     Po  190.0  135.0  197.0 nodata
85.     At nodata  127.0  202.0 nodata
86.     Rn nodata  120.0  220.0  145.0
87.     Fr nodata nodata  348.0 nodata
88.     Ra  215.0 nodata  283.0 nodata
89.     Ac  195.0 nodata nodata nodata
90.     Th  180.0 nodata nodata nodata
91.     Pa  180.0 nodata nodata nodata
92.      U  175.0 nodata  186.0 nodata
"""




"""
**A hexagonal lattice of cylinders with disorder**

We observe the suppression of higher order peaks with increasing disorder (Debye-Waller like factor)
"""

import jscatter as js

q = js.loglist(0.001, 5, 1000)
unitcelllength = 50
N = 5
rms = 1
domainsize = unitcelllength * N
hexgrid = js.sf.hex2DLattice(unitcelllength, N)
p = js.grace()
p.multi(2, 1)
cyl = js.formel.pDA(js.ff.cylinder, sig=0.01, parname='radius', q=q, L=50, radius=15, SLD=1e-4)
cyl.Y = js.formel.smooth(cyl, 20)
for i, rms in enumerate([1, 3, 10, 30], 2):
    hex = js.sf.latticeStructureFactor(q, lattice=hexgrid, rmsd=rms, domainsize=domainsize)
    p[1].plot(hex, li=[1, 2, i], sy=0)
    p[0].plot(hex.X, hex.Y * cyl.Y, li=[1, 3, i], sy=0, le='hex cylinders rms= ' + str(rms))
p[0].plot(cyl, li=[3, 2, 1], sy=0, le='cylinder formfactor')
# pretty up
p[1].yaxis(scale='n', label='I(Q)', max=10, min=0)
p[0].yaxis(scale='l', label='I(Q)')  # ,max=10000,min=0.01)
p[0].xaxis(scale='n', label='', min=0)
p[1].xaxis(scale='n', label=r'Q / A\S-1', min=0)
p[0].legend(x=0.6, y=60, charsize=0.8)
p[0].title('hex lattice of cylinders')
p[0].subtitle('increasing disorder rmsd')
p.save('hexagonalLatticeofCylinders1.png')



# -*- coding: utf-8 -*-
# written by Ralf Biehl at the Forschungszentrum Jülich ,
# Jülich Center for Neutron Science 1 and Institute of Complex Systems 1
#    Jscatter is a program to read, analyse and plot data
#    Copyright (C) 2015-2024  Ralf Biehl
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

"""
Collection of data tables. 

"""

__all__ = ['vdwradii', 'radiusBohr', 'Elements', 'felectron',
           'neutronFFgroup', 'xrayFFatomic', 'xrayFFatomicdummy', 'xrayFFgroupdummy', 'xrayFFgroup', 'Nscatlength',
           'aquasolventdensity', 'bufferDensityViscosity']

import os
import numpy as np
from scipy import constants
import io

path = os.path.realpath(os.path.dirname(__file__))

from ..dataarray import dataArray as dA
from ..dataarray import dataList as dL
from . import vanderWaalsRadii

#: Bohr radius in unit nm
radiusBohr = constants.physical_constants['Bohr radius'][0]*1e9  # in nm

#: van der Waals radii (mainly Bondi reference), classical vdW radii
#:
#: `<https://en.wikipedia.org/wiki/Atomic_radii_of_the_elements_(data_page)#Van_der_Waals_radius>`_
vdwradiiBondi = vanderWaalsRadii.getvdWdict(vanderWaalsRadii.tableofvdWRadii)
vdwradiiBondiA = {k.upper(): v * 10 for k, v in vdwradiiBondi.items()}

# see CRYSOL for details: for proteins volume determination use the vdW radii from
# Fraser, MacRae, Suzuki,(1978). Journal of Applied Crystallography, 11(6), 693–694.
# https://doi.org/10.1107/S0021889878014296
# as these result in nearly perfect volume occupancy in the protein SESVolume
vdwradii = vdwradiiBondi.copy()
vdwradii.update({'H': 0.107, 'C': 0.158, 'N': 0.084, 'O': 0.13, 'S': 0.168, 'P': 0.111})
vdwradii.update({'D': vdwradii['H']})


def _loglist(mini=0.1, maxi=5, number=100):
    return np.exp(
        np.r_[np.log((mini if mini!=0. else 1e-6)):np.log((maxi if maxi!=0 else 1.)):(number if number!=0 else 10)*1j])


# a global Qlist in 1/nm
QLIST = np.r_[0, _loglist(0.001, 0.1, 10), _loglist(0.1, 50, 40), _loglist(50, 100, 10)]

# Preparing Xray scattering factors for all elements
# according to International Tables for Crystallography (2006). Vol. C, ch. 4.3, pp. 259-262
# we use here parametrization of
# Rez, D., Rez, P. & Grant, I. (1994).
# Dirac–Fock calculations of X-ray scattering factors and
# contributions to the mean inner potential for electron scattering.
# Acta Cryst. A50, 481–497.
# which is valid up to 2.0 A^-1 (their units)
# to convert from their wavevector units to 1/nm we need their_q=our_q/4/pi/10
# scattering is per electron so xRayblength(q=0,symbol) results in charge Z
# to get scattering length in nm we need later to multiply with cross-section of electron

#: electron cross-section
felectron = 2.8179403267e-6   # in nm cross-section of electron

#: H2O solvent electron density in units e/nm^3 at 293.15K
solventElectrondensity = 333.685

solventXlengthDensity = felectron*solventElectrondensity

#: Xray scattering factors for all elements;
xRaycoefficients = {}

# load data for blength calculation here the coefficients for polynom
with open(path+'/'+'Xrayscattering_length.dat') as f:
    xRaytemp = f.readlines()
for line in xRaytemp:
    if line.startswith('#') or not line.strip():
        continue
    # first is number of electrons then we have the coefficients
    words = line.split()
    xRaycoefficients[words[0]] = np.asarray(words[1:], float)
XRC = xRaycoefficients
del xRaytemp


def xRayblength(q=None, symbol='', felectron=2.8179403267e-6):
    """
    Xray scattering amplitude for atoms.

    Parametrisation of Rez, D., Rez, P. & Grant, for felectron
    converts from electron density to 1/nm scattering length
    with felectron=1 we get the usual behaviour as  xRayblength(q=0,symbol)=Z
    CRYSOL uses felectron=1

    Parameters
    ----------
    q : array of float
        Wavevector
    symbol : string
        Symbol for element
    felectron : float
        Electron cross-section in nm = 2.8179403267e-6

    Notes
    -----
    see::
     Rez, D., Rez, P. & Grant, I. (1994).
     Dirac–Fock calculations of X-ray scattering factors and contributions to the mean inner potential for electron scattering.
     Acta Cryst. A50, 481–497.

    Rez et al.  is valid up to 2.0 A^-1 (their units)
    to convert from their wavevector units to 1/nm we need q_their=q_our/4/pi/10
    scattering is per electron so xRayblength(q=0,symbol) results in charge Z

    """
    if q is None:
        q=QLIST
    return felectron*(XRC[symbol][1]*np.exp(-XRC[symbol][2]*(q/4/np.pi/10)**2)+
                      XRC[symbol][3]*np.exp(-XRC[symbol][4]*(q/4/np.pi/10)**2)+
                      XRC[symbol][5]*np.exp(-XRC[symbol][6]*(q/4/np.pi/10)**2)+
                      XRC[symbol][7]*np.exp(-XRC[symbol][8]*(q/4/np.pi/10)**2))


# Formfactors of atoms and amino acids in X-ray scattering
#: X-ray scattering Volume of dummy atoms in nm**3  from van der Waals radii
XVolumes = {k: 4/3.*np.pi*v**3 for k, v in vdwradii.items()}
XVolumes['H2O'] = XVolumes['O'] + 2 * XVolumes['H']  # => 2*H+O


#: Dict of atomic xray formfactor for elements as dataArray with [q, coherent, incoherent]
xrayFFatomic = {}


#: Dict of dummy atomic xray formfactor (H2O at 293.15K) for elements as dataArray with [q, coherent, incoherent]
xrayFFatomicdummy = {}


###################################################
#: Dict of amino acid xray formfactors as dataArray with [q, coherent, incoherent]
xrayFFgroup = {}
xtemp = dL(path+'/'+'xrayFFgroup.fq')
for gx in xtemp:
    gx.getfromcomment('resname')
    gx.getfromcomment('mode')
    xrayFFgroup[gx.resname] = gx

# and the dummy for Xrays
xrayFFgroupdummy = {}
dtemp = dL(path+'/'+'xrayFFgroupdummy.fq')
for gd in dtemp:
    gd.getfromcomment('resname')
    gd.getfromcomment('mode')
    xrayFFgroupdummy[gd.resname] = gd

#: Dict of amino acid neutron formfactors as dataArray with [q, coherent, incoherent]
#: depends on actual deuteration
neutronFFgroup = {}
ntemp = dL(path+'/'+'neutronFFgroup.fq')
for gn in ntemp:
    gn.getfromcomment('resname')
    gn.getfromcomment('mode')
    neutronFFgroup[gn.resname] = gn

incXdat={}
with open(path+'/'+'compton.dat') as f:
    _temp = f.readlines()
for line in _temp:
    if line.startswith('#') or not line.strip():
        continue
    # name, number of electrons, coefficients
    # He 2  7.2391e-01  -2.1464e-01  9.1019e+00 1.5566e+01  0.01
    words = line.split()
    incXdat[words[0]] = np.asarray(words[1:], float)
del _temp

def _incoherentXray(fcoh):
    """
    Calculates the incoherent (inelastic scattering due to Compton effect) formfactors.


    Parameters
    ----------
    fcoh : dataArray
        Coherent formfactor of atom with attribute 'atomsymbol'.

    Returns
    -------
    array

    Notes
    -----
    To convert from wavevector units in [1]_ to 1/nm we need q_their=q_our/4/pi/10
    as it is defined as sin(theta)/lambda. Valid up to 1.2 1/A in their units.
    It is already multiplied by the electron scattering cross-section felectron**2.

    References
    ----------
    .. [1] A new analytic approximation to atomic incoherent X-ray scattering intensities.
           J. THAKKAR and DOUGLAS C. CHAPMAN, Acta Cryst. (1975). A31, 391

    .. [2] Analytic approximations for the incoherent X-ray intensities of the atoms from Ca to Am
           Acta Cryst. (1973). A29, 10-12, https://doi.org/10.1107/S0567739473000021
           G. Pálinkás

    .. [3] Atomic form factors, incoherent scattering functions, and photon scattering cross-sections
           J. H. Hubbell1, et al.J. Phys. Chem. Ref. Data 4, 471 (1975)
           http://dx.doi.org/10.1063/1.555523

    """
    a=0.9200
    b=0.176*4*np.pi*2.7769
    c=2.3266
    atomsymbol=fcoh.atomsymbol
    if atomsymbol.lower()=='h':
        # [3]_ equation 29
        Z=1
        fc=(fcoh.Y/felectron)**2
        finc=(1-fc)*felectron**2
        return finc
    Z=XRC[atomsymbol][0]
    if Z<36:
        # [1]_
        Z, a, b, c, d, err=incXdat[atomsymbol]
        X = fcoh.X/4/np.pi/10.  # -> sin(theta)/lambda units
        X2 = X*X
        finc = Z*(1-(1+a*X2+b*X2*X2)/(1+c*X2+d*X2*X2)**2)*felectron**2
    else:
        # [2]_
        X = fcoh.X/4/np.pi/10.  # -> sin(theta)/lambda units
        finc = Z*(1-a/(1+b*X/Z**(2/3.))**c)*felectron**2

    return finc


def G(q, r0, rm):
    # expansion factor for atomic formfactors for excluded Volumes
    return (r0 / rm) ** 3 * np.exp(-(4 * np.pi / 3) ** (3 / 2.) * np.pi * (q / 2 / np.pi) ** 2 * (r0 ** 2 - rm ** 2))


def ffxdummy(k, q, r0, rm):
    # calculation with subgroups like CRYSOL
    # ff = lambda key,q: G(q,r0,rm) * Xld * XVolumes(key) * np.exp(-np.pi*q**2 * XVolumes(key)**(2/3.))
    # atomic base calculation rm==r0
    # q/2/np.pi because of different definition of fourier transform
    XV = XVolumes[k]
    if r0 != rm:
        return G(q, r0, rm) * solventXlengthDensity * XV * np.exp(-np.pi * (q / 2 / np.pi) ** 2 * XV ** (2. / 3.))
    else:
        return solventXlengthDensity * XV * np.exp(-np.pi * (q / 2 / np.pi) ** 2 * XV ** (2. / 3.))


def _calcXrayFFatomic(q, r0=1, rm=1, xraywavelength=0.15418):
    """
    Predefines dictionary of atomic scattering amplitude and incoherent formfactors for X-rays.

    Parameters
    ----------
    q : array
        Wavevectors in 1/nm
    r0,rm=1 : float
        see CRYSOL
    xraywavelength : float, default 0.15418
        Xray wavelength in nm, default is K_alpha of copper as used in SAXSPACE

    Returns
    -------
    dataArray
        columns [q, coherent, incoherent2] in units [1/nm, nm, nm**2]

    Notes
    -----
    * According to International Tables for Crystallography (2006). Vol. C, ch. 4.3, pp. 259-262
      we use here parametrisation of [1]_. hydrogen formfactor fh = lambda q:16/(4+q**2*radiusBohr**2)**2
    * The standard deviation is about 1.4% or 1% for proteins with Molecular weight >30kDa.
    * It depends on the protein Volume which is adjusted to result in a density 1.37 g/ml.

    References
    ----------
    .. [1] Dirac–Fock1 calculations of X-ray scattering factors and contributions to the mean inner potential for
           electron scattering.
           Rez, D., Rez, P. & Grant, I. .Acta Cryst. A50, 481–497.(1994)

    """
    # use here global variables
    global xrayFFatomic, xrayFFatomicdummy

    # pol correction is <0.97 for q<10 1/nm should be included for larger q
    theta = lambda q: 2*np.arcsin(q*xraywavelength/4./np.pi)
    unpol = lambda theta: 0.5*(1+np.cos(theta)**2)

    # calc if not already calculated for same qlist
    if 'H' not in xrayFFatomic or not np.all(xrayFFatomic['H'].X == q):
        # hydrogen formfactor fh
        fh = lambda q: felectron*16 / (4+q**2*radiusBohr**2)**2

        # H atomic formfactor for X-rays
        xrayFFatomic['H'] = dA(np.array([q, fh(q), q*0]), XYeYeX=[0, 1])
        xrayFFatomic['H'].atomsymbol = 'H'
        xrayFFatomic['H'][2] = _incoherentXray(xrayFFatomic['H'])
        xrayFFatomic['H'].comment.append('columns= q; coherent_amplitude; incoherent_formfactor')
        xrayFFatomic['H'].columnname = 'q; coh; inc2'
        xrayFFatomic['D'] = xrayFFatomic['H']
        xrayFFatomic['D'].atomsymbol = 'D'

        for key in xRaycoefficients:
            # tabulated formfactors for not H atoms
            FF = xRayblength(q, key)
            xrayFFatomic[key] = dA(np.array([q, FF, np.zeros_like(q)]), XYeYeX=[0, 1])
            xrayFFatomic[key].atomsymbol = key
            xrayFFatomic[key].comment.append('columns= q; coherent_amplitude; incoherent_formfactor')
            xrayFFatomic[key].columnname = 'q; coh; inc2'
            xrayFFatomic[key][2] = _incoherentXray(xrayFFatomic[key])

        # calculate dummy formfactors with electron density of solvent at 293.15K
        # and scattering length of felectron
        for key in XVolumes:
            xrayFFatomicdummy[key] = dA(np.array([q, ffxdummy(key, q, r0, rm), q*0]))
            if key == 'H2O':
                xrayFFatomicdummy[key][2] = xrayFFatomic['H'][2]*2 + xrayFFatomic['O'][2]


def getxrayFFatomic(atom):
    """
    Coherent scattering amplitudes and incoherent formfactors for X-rays.

    As Xrays are not sensitive to deuterium exchange we can base it on atom type.
    Dummy atoms need atom.electrondensity and atom.Xvolume as created with makeDummyMolecule.

    Parameters
    ----------
    atom : atom, str
        atom or atom type as 'C','O','H','D','Mg'
        see Xrayscattering_length.dat or atom.type

    Returns
    -------
    dataArray
        columns [q, coherent scattering amplitude, incoherent fomfactor] in units [1/nm, nm, nm**2] per molecule.

    Notes
    -----
    Definitions::

     q          = 4pi/lambda*sin(theta/2)  wavevector in units 1/nm
     coherent   scattering intensity Icoh = getxrayFFatomic(atom).Y**2
     incoherent scattering intensity Iinc = getxrayFFatomic(atom)[2]

    - Coherent:  parametrisation of [1]_ for coherent scattering amplitudes.
    - Incoherent: Formfactors are defined in [2,3,4]_ for incoherent Compton scattering.

    | hydrogen formfactor fh
    | fh=lambda q:felectron*16/(4+q**2*radiusBohr**2)**2
    | hydrogen incoherent scattering
    | finc=(1-(fh/felectron)**2)*felectron**2

    Dummy atom model a la CRYSOL for excluded volume scattering in getxrayFFatomicdummy.
    This model is also used for dummy molecule modeling of Xrays.

    There is no Debye-Waller-factor included as in equation 9 of [5]_.
    This would need a B factor from pdb structure
    and add a factor exp(-B_j*q^2/(16*pi^2)) to the  xRayblength(q,key) individual for each atom_j

    To convert from their wavevector units ([1,2,3,4]_) to 1/nm we need q_their=q_our/4/pi/10
    as it is defined as sin(theta)/lambda. Valid up to 1.2 1/A in their units.

    References
    ----------
    .. [1] Dirac–Fock calculations of X-ray scattering factors and contributions to the mean inner potential for electron scattering.
           Rez, D., Rez, P. & Grant, I. (1994).
           Acta Cryst. A50, 481–497.

    .. [2] A new analytic approximation to atomic incoherent X-ray scattering intensities.
           J. THAKKAR and DOUGLAS C. CHAPMAN, Acta Cryst. (1975). A31, 391

    .. [3] Analytic approximations for the incoherent X-ray intensities of the atoms from Ca to Am
           Acta Cryst. (1973). A29, 10-12, https://doi.org/10.1107/S0567739473000021
           G. Pálinkás

    .. [4] Atomic form factors, incoherent scattering functions, and photon scattering cross-sections
           J. H. Hubbell1, et al.J. Phys. Chem. Ref. Data 4, 471 (1975)
           http://dx.doi.org/10.1063/1.555523

    .. [5] Softwax: a computational tool for modelling wide range xray....
           Bardhan et al.J. applied Crystallography 42,932-943 (2009)

    """
    global xrayFFatomic, xrayFFatomicdummy
    if xrayFFatomic=={}:
        _calcXrayFFatomic(QLIST)

    try:
        type=atom.type
    except:
        type=atom
    try:
        # predefined in calcXrayFFatomic with real atoms
        return xrayFFatomic[type]
    except:
        raise KeyError('atom not found in xrayFFatomic.', type)


def getxrayFFatomicdummy(atom):
    """
    returns dummy atomic scattering amplitude for X-rays with scattering length of solvent (water)

    see getxrayFFatomic
    """
    global xrayFFatomic, xrayFFatomicdummy
    if xrayFFatomic == {}:
        _calcXrayFFatomic(QLIST)
    try:
        type = atom.type
    except:
        type = atom

    if isinstance(atom, str) and atom.lower() == 'h2o':
        return xrayFFatomicdummy['H2O']
    else:
        return xrayFFatomicdummy[type]


# ask once to precalculate all xrayFFatomic during importing this module
getxrayFFatomicdummy('C')
getxrayFFatomic('C')

# --------------------------------------------------
# load table with neutron scattering cross-sections
with io.open(os.path.join(path, 'Neutronscatteringlengthsandcrosssections.html')) as _f:
    Nscatdat=_f.readlines()

#: Dictionary with  neutron scattering length for elements as [b_coherent, b_incoherent].
#: units nm
#: from `<http://www.ncnr.nist.gov/resources/n-lengths/list.html>`_
Nscatlength={}

Hnames={'1H': 'h', '2H': 'd', '3H': 't'}
for line in Nscatdat[115+4:486+4]:
    words=[w.strip() for w in line.split('<td>')]
    if words[1] in Hnames.keys():
        Nscatlength[Hnames[words[1]]]=[float(words[3])*1e-6, np.sqrt(float(words[6])/4/np.pi*1e-10)]
    elif words[1][0] not in '0123456789':
        # noinspection PyBroadException
        try:
            Nscatlength[words[1].lower()]=[float(words[3])*1e-6, np.sqrt(float(words[6])/4/np.pi*1e-10)]
        except:
            # noinspection PyBroadException
            try:
                Nscatlength[words[1].lower()]=[complex(words[3])*1e-6, np.sqrt(float(words[6])/4/np.pi*1e-10)]
            except:
                Nscatlength[words[1].lower()]=[-0, -0]
del words

#  [Z, mass, b_coherent, b_incoherent, name]
#: Elements Dictionary
#: with: { symbol : (e- number; mass; neutron coherent scattering length, neutron incoherent scattering length, name)};
#: units amu for mass and nm for scattering length
Elements={}

# load periodic table perhaps later more of this
with io.open(os.path.join(path, 'elementsTable.dat')) as _f:
    for ele in _f.readlines():
        if ele[0] == '#': continue
        z, symbol, name, mass=ele.split()[0:4]
        try:
            Elements[symbol.lower()]=(int(z),
                                      float(mass),
                                      Nscatlength[symbol.lower()][0],
                                      Nscatlength[symbol.lower()][1],
                                      name)
        except KeyError:
            pass
del z, symbol, name, mass

#: water molecular volume in units nm³ from density
water_vol=1e8**3/(1000/18.*constants.N_A)


###################################################

#: Hydrophobicity of aminoacids
#: Kyte J, Doolittle RF., J Mol Biol. 1982 May 5;157(1):105-32.
aaHydrophobicity = {}
with open(path+'/'+'hydrophobicity.dat') as f:
    _temp = f.readlines()
for line in _temp:
    if line.startswith('#') or not line.strip():
        continue
    words = line.split()
    aaHydrophobicity[words[0].lower()] = np.asarray(words[1:5], float)
del _temp



# load table with density parameters according to
# Densities of binary aqueous solutions of 306 inorganic substances
# P. Novotny, O. Sohnel J. Chem. Eng. Data, 1988, 33 (1), pp 49–55 DOI: 10.1021/je00051a018
aquasolventdensity = {}
with open(os.path.join(path, 'aqueousSolutionDensitiesInorganicSubstances.txt')) as _f:
    for ele in _f.readlines():
        if ele[0] == '#': continue
        # substance A*10^-2 -B*10 C*10^3 -D E*10^2 -F*10^4 st*10^2 t Cmax-
        aname, A, B, C, D, E, F, s, Trange, concrange = ele.split()[0:10]
        aquasolventdensity[aname.lower()] = (float(A) * 1e2,
                                              -float(B) * 1e-1,
                                              float(C) * 1e-3,
                                              -float(D) * 1e0,
                                              float(E) * 1e-2,
                                              -float(F) * 1e-4,
                                              float(s) / 100.,
                                              Trange,
                                              concrange)

aquasolventdensity['c4h11n1o3'] = (0.0315602, 0.708699)  #: TRIS buffer density    DOI: 10.1021/je900260g
aquasolventdensity['c8h19n1o6s1'] = (0.0774654, 0.661610)  #: TABS buffer density    DOI: 10.1021/je900260g
del aname, A, B, C, D, E, F, s, Trange, concrange, ele


bufferDensityViscosity = {}
with io.open(os.path.join(path, 'bufferComponents.txt'), 'r') as _f:
    for ele in _f.readlines():
        if ele[0] == '#': continue
        # substance
        name, dc0, dc1, dc2, dc3, dc4, dc5, vc0, vc1, vc2, vc3, vc4, vc5, unit, crange = ele.split()
        temp = [float(ss) for ss in [dc0, dc1, dc2, dc3, dc4, dc5, vc0, vc1, vc2, vc3, vc4, vc5]]
        bufferDensityViscosity[name.lower()] = tuple(temp) + (unit, crange)
        # except:
        #    pass
del ele, name, dc0, dc1, dc2, dc3, dc4, dc5, vc0, vc1, vc2, vc3, vc4, vc5, unit, crange, temp






# -*- coding: utf-8 -*-
# written by Ralf Biehl at the Forschungszentrum Jülich ,
# Jülich Center for Neutron Science 1 and Institute of Complex Systems 1
#    Jscatter is a program to read, analyse and plot data
#    Copyright (C) 2015-2024  Ralf Biehl
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

"""
This module contains the scattering functions.

"""

import os
import re
import sys
import subprocess
import shutil
import tempfile
import numbers

import numpy as np
from scipy import constants as co
from scipy.spatial.transform import Rotation as Rot

import MDAnalysis as mda
from MDAnalysis.topology.guessers import guess_masses
import MDAnalysis.transformations as trans

from .. import formel
from .. import data
from .mda import scatteringUniverse


try:
    import pymol2
except ImportError:
    pymol2 = False


__all__ = ['runHydropro', 'readHydroproResult', 'createWaterBox']


def is_float(s):
    try:
        float(s)
        return True
    except ValueError:
        return False


def is_dataline(words):
    """
    Test if line words starts with float.
    wf : list of strings

    """
    try:
        return is_float(words[1]) and is_float(words[0])
    except IndexError:
        return False


def runHydropro(atomgroup, outfile=None, type=1, T=20, visc=0, AER=None, NSIG=-1, SIGMAX=4.0, SIGMIN=3.0,
                MW=None, soldensity=1.11, solvent='d2o', showoutput=0, hydropro=None):
    """
    Diffusion tensor, Dtrans, Drot, S and more from PDB structure using HYDROPRO.

    HYDROPRO computes the hydrodynamic properties of rigid macromolecules from PDB structures.
    This wrapper writes the input file for HYDROPRO [1]_  hydropro.dat, runs HYDROPRO and reads the result file.
    The hydropro executable named 'hydropro10-lnx.exe' needs to be in the PATH. Download from [2]_.

    Use :py:func:`~.readHydroproResult` to read an existing result file.

    Parameters
    ----------
    atomgroup : atomgroup, string
        A MDAnalysis atomgroup (e.g. uni.select_atoms('protein') or the filename of a PDB structure.
        If the filename contains a prepended path this is used as working directory wor the output.
    outfile : string, default None
        Output filename to use, by default a name is generated from residue names and Rg.
    type : 1,2,4
        - 1 -- Atomic-level primary model, shell calculation, default and only for all atom model
        - 2 -- Residue-level primary model, shell calculation, default residue model (Ca only)
        - 4 -- Residue-level primary model, bead calculation, (Ca only)
    T : float default 20
        Temperature in °C.
    visc : float , default 0
        Viscosity in poise=10*Pa*s. H2O@20C is 0.01 Poise.
        0 means calc from temperature T for solvent.
    AER : float, default depends on type
        The value of the hydrodynamic radius, in Å. Default values a described in [1]_ and should not be changed.

        - type 1 default to 2.9 A
        - type 2 default to 4.8 A
        - type 4 default to 6.1 A
    NSIG : int, default -1
        NSIG (INTEGER) Number of values of the radius of the mini bead.
        -1 autodetermine SIGMIN SIGMAX
    SIGMAX : float, default 4.0
        Lowest value of sigma, the mini bead radius
    SIGMIN : float, default 3.0
        Highest value of sigma, the mini bead radius
    MW : float, default None
        Molecular weight; if None calculated from universe
    soldensity : float,default 1.1
        Solvent density  1.1 is d2o
    solvent : 'd2o' or 'h2o'
        Solvent
    showoutput : 0
        Show output of hydropro 0 ->minimal; None ->No output; other ->full output
    hydropro : string, default 'hydropro10-lnx.exe' or 'hydropro10-msd.exe'
        Filename of the hydropro executable in PATH.
        For Windows or Linux the default is tested.
        Change this if you use a different name.

    Returns
    -------
    dict with results of 6x6 mobility matrix with 4 3x3 matrices
        - 'DTT' : translational 3x3 matrix, units nm^2/ps
        - 'DRR' : rotational 3x3 matrix, units 1/ps
        - 'DTR' ='DRT'^T : translational rotational coupling, units nm/ps
        -  other keys with units as given in values

    Examples
    --------
    ::

     import jscatter as js
     uni = js.bio.scatteringUniverse('3rn3',addHydrogen=False)
     H = js.bio.runHydropro(uni.atoms)
     uni.qlist = np.r_[0.01:2:0.1]
     D = js.bio.diffusionTRUnivTensor(uni,DTT=H['DTT'],DRR=H['DRR'],DRT=H['DRT'])


    References
    ----------
    .. [1] Prediction of hydrodynamic and other solution properties of rigid proteins
           from atomic- and residue-level models
           A. Ortega, D. Amorós, J. Garc1a de la Torre,
           Biophys. J. 101, 892-898 (2011)
    .. [2] http://leonardo.inf.um.es/macromol/programs/hydropro/hydropro.htm


    """
    # test which hydropro
    if hydropro is None:
        if sys.platform.startswith('linux'):
            hydropro = 'hydropro10-lnx.exe'
        elif sys.platform.startswith('win'):
            hydropro = 'hydropro10-msd.exe'
        else:
            raise NotImplementedError('hydropro executable only for windows and linux.')
    hydroproexe = shutil.which(hydropro)
    if not hydroproexe:
        raise FileNotFoundError('No hydropro executable found in PATH.')

    path = os.getcwd()
    # noinspection PyAugmentAssignment
    if hasattr(atomgroup, 'write'):
        if outfile is None:
            # generate a specific name
            aminos = [mda.lib.util.convert_aa_code(res)
                      for res in atomgroup.select_atoms('protein').residues[:5].resnames]
            nucleic = [r[0] for r in atomgroup.select_atoms('nucleic').residues[:5].resnames]
            pdbfile = ''.join(aminos+nucleic)[:5]
            if not pdbfile: pdbfile = 'compl'
            pdbfile += f'{atomgroup.total_mass():.0f}'
        else:
            pdbfile=outfile
        Rg = atomgroup.radius_of_gyration()
        pdbfile = pdbfile+f'Rg{10*Rg:.0f}T{T:.0f}'
        pdbfile = pdbfile[:20]

        atomgroup.write(pdbfile+'.pdb')

        if np.all(atomgroup.atoms.names == 'CA'):
            # for Calpha models
            if MW is None:
                MW = atomgroup.n_atoms * 109
            if type !=4:
                type = 2
                AER = 4.8  # in A for Ca models
            else:
                AER = 6.1  # in A for Ca models
        else:
            if MW is None:
                MW = atomgroup.total_mass()
            type = 1
            AER = 2.9  # in A for atomic models

    elif os.path.isfile(atomgroup+'.pdb') or os.path.isfile(atomgroup):
        path, file = os.path.split(atomgroup)
        pdbfile, _ = os.path.splitext(file)
        if AER is None:
            if type == 1:
                AER = 2.9
            elif type == 2:
                AER = 4.8
            elif type == 4:
                AER = 6.1
    else:
        raise ValueError('Input is not atomgroup nor pdb file!')

    if len(pdbfile) > 38:
        raise NameError('PDB filename to long. Should be not more than 38 char.')

    # generate the input file
    zeilen = [pdbfile,
              pdbfile,
              pdbfile + '.pdb',
              f'{type:.0f},                  !Type of calculation ',
              f'{AER:.1f},                   !AER, radius of the atomic elements',
              f'{NSIG:.1f},                  !NSIG ANzahl R zwischen min max']
    if NSIG != -1:
        zeilen.append(f'{SIGMIN:.1f},              !Minimum radius of beads in the shell (SIGMIN)')
        zeilen.append(f'{SIGMAX:.1f},              !Maximum radius of beads in the shell (SIGMAX)')
    zeilen.append(f'{T:.1f},           !T (temperature, C)')
    if visc == 0:
        visc = formel.viscosity(mat=solvent, T=273.15+T) * 10
    zeilen.append(f'{visc:.4f},                !ETA ')
    zeilen.append(f'{MW:.1f},                  !RM (Molecular weigth)')
    zeilen.append('-1.0,                       !partial specific volume, cm3/g')
    zeilen.append(f'{soldensity:.2f},          !Solvent density, g/cm3')
    zeilen.append('0,               !NQ Number of values of Q')
    zeilen.append('0,               !Number of intervals for the distance distribution')
    zeilen.append('0,               !Number of trials for MC calculation of covolume')
    zeilen.append('1                 !IDIF=1 (yes) for full diffusion tensors')
    zeilen.append('*                 !End of file')

    hydroprodatpath = os.path.join(path, 'hydropro.dat')
    with open(hydroprodatpath, 'w') as f:
        f.writelines(' \n'.join(zeilen))

    hydroproexe = shutil.which(hydropro)
    if not hydroproexe:
        raise FileNotFoundError('No hydropro executable found in PATH.')

    p = subprocess.run(hydroproexe, shell=True, capture_output=True, cwd=path)

    if p.stderr != '':
        for line in p.stderr.split(b'\n'):
            print('hydropro_std_err>', line)

    if showoutput:
        for line in p.stdout.split(b'\n'):
            print('hydropro>', line)

    elif showoutput is not None:
        linestoshow = [b'poise', b'Molecular weight', b'Translational']
        for line in p.stdout.split(b'\n'):
            if any(word in line for word in linestoshow):
                print('hydropro>', line)
    #
    result = readHydroproResult(os.path.join(path, pdbfile)+'-res.txt')

    return result


def readHydroproResult(file='.res'):
    """
    Reads the result file of HYDROPRO.

    Parameters
    ----------
    file          filename in dir or full path without dir

    Returns
    -------
    dict with results of 6x6 mobility matrix with 4 3x3 matrizes
        - 'DTT' : translational 3x3 matrix, units nm^2/ps
        - 'DRR' : rotational 3x3 matrix, units 1/ps
        - 'DTR' ='DRT'^T : tranlational rotational coupling, units nm/ps
        -  other keys with units as given in values


    """
    with open(file) as f:
        zeilen = f.readlines()

    H = []
    result = {}
    empty = re.compile(r'^\s*$')
    for zeile in zeilen:
        if empty.match(zeile):
            continue
        words = zeile.split()
        if is_dataline(words):
            H.append([float(w) for w in words])
        elif ':' in zeile:
            try:
                words = zeile.split(':')
                key = words[0].strip().replace(' ','_')
                vals = words[1].split()
                if is_float(vals[0]):
                    val = float(vals[0])
                    result[key] = (val,) + tuple(vals[1:])
            except:
                pass
    aH = np.array(H)
    result['DTT'] = aH[:3, :3] * 1.e+2
    result['DTR'] = aH[:3, 3:] * 1.e-5
    result['DRT'] = aH[3:, :3] * 1.e-5
    result['DRR'] = aH[3:, 3:] * 1.e-12

    return result


def savePymolpng(atomgroup, fname, rotate=[]):
    """
    Save png image of current configuration using Pymol.

    A simplified method to make png images.
    For more control use this function as a template and adopt it to your needs.
    Pymol needs to be installed (version >2.4) that `pymol2` can be imported.
    `pymol2` provides an additional interface to PyMol.


    Parameters
    ----------
    atomgroup : atomgroup
        Atomgroup to show in plot.
    fname : string
        Filename
    rotate : list 3xfloat
        Angles to rotate around ['x','y','z'] axis.
        Missing values are zero.

    """
    if isinstance(rotate,numbers.Number):
        rotate = [rotate]
    if pymol2:
        with pymol2.PyMOL() as p1:
            with tempfile.TemporaryDirectory() as tdir:
                name = os.path.splitext(atomgroup.universe.filename)[0] + '.pdb'
                tfile = os.path.join(tdir, name)
                atomgroup.write(tfile)
                p1.cmd.load(tfile)

                for ax, angle in zip(['x','y','z'], rotate):
                    p1.cmd.rotate(ax, angle, 'all')

                # set colors
                p1.cmd.color('red', 'ss h')
                p1.cmd.color('yellow', 'ss s')
                p1.cmd.color('blue', 'ss l+')
                p1.cmd.set('cartoon_nucleic_acid_color', 'yellow')
                p1.cmd.set('cartoon_ladder_color', 'green')
                p1.cmd.set('cartoon_discrete_colors',1)
                p1.cmd.set('cartoon_ring_mode', 1)
                p1.cmd.color('green', '(resn DA+DC+DG+DT)')

                # make png
                p1.cmd.png(fname, width=600, height=600, dpi=-1, ray=1)

    else:
        print('Pymol needed but Pymol not installed or old version.')


h2o = np.array([[ 0,        0,       0      ],   # oxygen
                [ 0.95908, -0.02691, 0.03231],   # hydrogen
                [-0.28004, -0.58767, 0.70556]])  # hydrogen


def createWaterBox(size, density=None, temperature=293.15, skip=321):
    """
    Create a MDAnalysis universe filled with water.

    Water positions are on a pseudorandom grid (ses :py:`~.formel.randomPointsInCube` ).

    Parameters
    ----------
    size : float
        Edge length of the universe box in units A.
    density : float
        Density in the waterbox in g/ml for pure H2O.
        If None H2O density at the given temperature is used.
        D2O density is a bit diffferent than H2O.
    temperature :
        Temperature of universe in K.
    skip : integer
        Skip this number of points in Halton sequence to get random configurations.
        For same number we get always the same configuration.

    Returns
    -------
    MDAnalysis universe quasi random water molecules in random orientation.
    Solvent ``resname='HOH"``

    To create scattering universe
    ::

     import jscatter as js
     u = js.bio.createWaterBox(50)



    """
    if density is None:
        density = formel.waterdensity('1H2O1', T=temperature)
    wmass = 2 * data.Elements['h'][1] + data.Elements['o'][1]

    n_residues = int((size/10)**3 / wmass * co.N_A/1e24*1000 * density)
    n_atoms = n_residues * 3
    resindices = np.repeat(range(n_residues), 3)
    segindices = [0] * n_residues
    sol = mda.Universe.empty(n_atoms,
                         n_residues=n_residues,
                         atom_resindex=resindices,
                         residue_segindex=segindices,
                         trajectory=True)

    sol.add_TopologyAttr('name', ['O', 'H1', 'H2'] * n_residues)
    sol.add_TopologyAttr('type', ['O', 'H', 'H'] * n_residues)
    sol.add_TopologyAttr('resname', ['HOH'] * n_residues)
    sol.add_TopologyAttr('resid', list(range(1, n_residues + 1)))
    sol.add_TopologyAttr('segid', ['SOL'])
    sol.add_TopologyAttr('mass', guess_masses(sol.atoms.types))

    op = formel.randomPointsInCube(n_residues, skip) * size
    coord_array = (op[:, None] + h2o).reshape(n_residues * 3, 3)
    sol.atoms.positions = coord_array
    sol.dimensions = [size, size, size, 90, 90, 90]

    bonds = []
    for o in range(0, n_atoms, 3):
        bonds.extend([(o, o+1), (o, o+2)])
    sol.add_TopologyAttr('bonds', bonds)

    for w, a, b in zip(sol.residues, np.random.rand(n_residues)*2*np.pi, np.random.rand(n_residues)*2*np.pi):
        com = w.atoms[0].position
        v = com - w.atoms[1].position
        R = Rot.from_rotvec(v/np.linalg.norm(v) * a)
        w.atoms[2].position = R.as_matrix() @ (w.atoms[2].position - com) + com
        v = com - w.atoms[2].position
        R = Rot.from_rotvec(v/np.linalg.norm(v) * b)
        w.atoms[1].position = R.as_matrix() @ (w.atoms[1].position - com) + com

    water = sol.select_atoms('not protein')
    transforms = [trans.wrap(water, compound='fragments')]

    sol.trajectory.add_transformations(*transforms)

    boxvolume = mda.lib.mdamath.box_volume(sol.dimensions)  # in A³
    # waterdensity = watermass / co.N_A / boxvolume * 1e24  # ~0.997
    numdensity = sol.residues.n_residues / boxvolume * 1000

    uni = scatteringUniverse(sol, guess_bonds=False)

    # set numdensity to exact values dependent on total mass
    uni.setSolvent(solvent='1H2O1', numDensity=numdensity)

    return uni








"""
Not yet working
number density wrong

A complex **example with solvent** based on an example from
`MDAnalysis Trajectory transformations
<https://docs.mdanalysis.org/stable/documentation_pages/trajectory_transformations.html#>`_
For MD simulation in a box of solvent we have to take care that the objects of interest, here a protein,
is not placed at the border of the simulation box.


 - The first issue is here that the trajectory contains dummy atoms for the solvent center of mass
   additional to the solvent atoms. We have to remove these.
 - The second issue is the position of the protein. The transformation shifts coordinates to have the protein
   in the box and a surrounding box full of solvent with not broken solvent molecules.
 - Then we compare scattering with and without explicit solvent.

"""


from MDAnalysisTests.datafiles import TPR, XTC
import jscatter as js
import numpy as n
import MDAnalysis as mda
import MDAnalysis.transformations as trans

# the original types are like 'opls_113' so first determine types from names, then reassign missing types
# and ignore sol for guess_bonds
u = mda.Universe(TPR, XTC)

# remove zero mass dummy atoms (solvent center of mass) present in topology
uu = mda.Merge(u.select_atoms('mass 0.1 to 100'))

# now create the scattering universe but guess_bonds not_for_solvent
utrp = js.bio.scatteringUniverse(uu, assignTypes={"from_names": 1}, guess_bonds='not segid seg_1_SOL')
# add original dimension
utrp.dimensions = u.dimensions

# check the universe, the protein is across the border
utrp.view(frames="all")

# add a transform that puts the protein into the center of the box and wrap the solvent into the new box
# for details see  MDAnalysis Trajectory transformations
protein = utrp.select_atoms('protein')
not_protein = utrp.select_atoms('not protein')
transforms = [trans.unwrap(protein),
              trans.center_in_box(protein, wrap=True),
              trans.wrap(not_protein, compound='fragments')]

utrp.trajectory.add_transformations(*transforms)
utrp.view(frames="all")

# compare protein scattering with and without solvent
utrp.qlist = js.loglist(0.1, 35, 300)
utrp.setSolvent('1h2o1')
Sqimplicit = js.bio.xscatIntUniv(utrp.select_atoms('protein'))
selsolvent = '(resname SOL) and (not type H) and not protein and (not around 6 protein)'
Sqexplicit = js.bio.xscatIntUniv(utrp.atoms, getVolume='box', selectSolvent=selsolvent)

boxvolume = utrp.boxVolume
nsol = utrp.select_atoms('resname SOL').n_residues
numberdensity = nsol / (boxvolume - Sqimplicit.SESVolume) * 1000  # 1/nm³
utrp.numDensitySol  # from setSolvent

p = js.grace()
p.yaxis(scale='log', label='I(Q)')
p.xaxis(scale='log', label='Q / nm\S-1')
p.plot(Sqexplicit, le='with solvent')
p.plot(Sqimplicit, le='only protein')
p.legend()
